from unittest.mock import AsyncMock

import pytest

from nexios.websockets import WebSocket, WebSocketConsumer


@pytest.fixture
def mock_websocket():
    scope = {
        "type": "websocket",
        "path": "/ws",
        "headers": [],
        "client": [],
        "server": [],
        "app": None,
        "extensions": {},
        "scope": {},
        "root_path": "",
        "url_scheme": "ws",
        "query_string": "",
        "headers_list": [],
        "cookies": {},
        "session": None,
        "user": None,
        "user_state": {},
        "user_state_policy": None,
    }
    receive = AsyncMock()
    send = AsyncMock()
    return WebSocket(scope, receive, send)


async def test_websocket_connection(mock_websocket):
    mock_websocket._receive.return_value = {"type": "websocket.connect"}
    await mock_websocket.accept()

    mock_websocket._send.assert_called_with(
        {"type": "websocket.accept", "subprotocol": None, "headers": []}
    )
    assert mock_websocket.application_state.value == 1  # CONNECTED


async def test_websocket_send_receive(mock_websocket):
    mock_websocket._receive.return_value = {"type": "websocket.connect"}
    await mock_websocket.accept()

    await mock_websocket.send_text("Hello")
    mock_websocket._send.assert_called_with({"type": "websocket.send", "text": "Hello"})

    mock_websocket._receive.return_value = {
        "type": "websocket.receive",
        "text": "World",
    }
    text = await mock_websocket.receive_text()
    assert text == "World"


async def test_websocket_json(mock_websocket):
    mock_websocket._receive.return_value = {"type": "websocket.connect"}
    await mock_websocket.accept()

    data = {"message": "Hello"}
    await mock_websocket.send_json(data)
    mock_websocket._send.assert_called_with(
        {"type": "websocket.send", "text": '{"message":"Hello"}'}
    )

    mock_websocket._receive.return_value = {
        "type": "websocket.receive",
        "text": '{"reply":"World"}',
    }
    received = await mock_websocket.receive_json()
    assert received == {"reply": "World"}


async def test_websocket_close(mock_websocket):
    # Test normal closure
    await mock_websocket.close(code=1000, reason="Normal closure")
    mock_websocket._send.assert_called_with(
        {"type": "websocket.close", "code": 1000, "reason": "Normal closure"}
    )


# Test WebSocket Consumer
class TestConsumer(WebSocketConsumer):
    encoding = "json"

    async def on_connect(self, websocket):
        await super().on_connect(websocket)
        await self.join_group("test_group")

    async def on_receive(self, websocket, data):
        await self.broadcast(data, group_name="test_group")



import re
from functools import wraps
from typing import Any, Awaitable, Callable

from nexios.http import Request, Response
from nexios.types import HandlerType


def use_for_route(route: str) -> None:
    if route.endswith("/*"):
        route = route[:-2]
        route = f"^{route}/.*$"
    else:
        route = f"^{route}$"

    def decorator(func: HandlerType) -> Any:
        @wraps(func)
        async def wrapper_func(
            request: Request,
            response: Response,
            call_next: Callable[..., Awaitable[Response]],
        ) -> Any:
            if re.match(route, request.url.path):
                return await func(request, response, call_next)  # type:ignore
            else:
                return await call_next()

        @wraps(func)
        async def wrapper_klass(
            self: Any,
            request: Request,
            response: Response,
            call_next: Callable[..., Awaitable[Response]],
        ) -> Any:
            if re.match(route, request.url.path):
                return await func(self, request, response, call_next)  # type:ignore
            else:
                return await call_next()

        if func.__name__ == "__call__":
            return wrapper_klass
        else:
            return wrapper_func

    return decorator  # type:ignore



import warnings
import nexios.status

def test_http_status_constants():
    assert nexios.status.HTTP_200_OK == 200
    assert nexios.status.HTTP_404_NOT_FOUND == 404
    assert nexios.status.HTTP_500_INTERNAL_SERVER_ERROR == 500

def test_websocket_status_constants():
    assert nexios.status.WS_1000_NORMAL_CLOSURE == 1000
    assert nexios.status.WS_1011_INTERNAL_ERROR == 1011

def test_deprecated_ws_status_warns():
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        val = getattr(nexios.status, "WS_1004_NO_STATUS_RCVD")
        assert val == 1004
        assert any("deprecated" in str(warn.message) for warn in w) 


import os
import shutil
import subprocess
import time
from datetime import datetime

import requests


def clean_previous_results():
    """Clean up previous results and plots"""
    # Clean results directory
    results_dir = "results"
    if os.path.exists(results_dir):
        shutil.rmtree(results_dir)
    os.makedirs(results_dir)

    # Clean plots directory
    plots_dir = "plots"
    if os.path.exists(plots_dir):
        shutil.rmtree(plots_dir)
    os.makedirs(plots_dir)

    # Remove previous summary files
    for file in os.listdir("."):
        if file.startswith("benchmark_summary_") and file.endswith(".csv"):
            os.remove(file)


def wait_for_server(url, max_retries=30, retry_delay=1):
    """Wait for server to be ready"""
    for i in range(max_retries):
        try:
            response = requests.get(url)
            if response.status_code == 200:
                return True
        except requests.exceptions.ConnectionError:
            pass
        time.sleep(retry_delay)
    return False


def run_server(server_file, port):
    return subprocess.Popen(
        ["uv", "run", server_file], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )


def run_locust_test(test_file, framework_name):
    # Run locust for 1 minute with 10 users
    # Each endpoint has weight 1, so with 4 endpoints, we need 75 requests per endpoint
    # to get 300 total requests (75 * 4 = 300)
    cmd = [
        "locust",
        "-f",
        test_file,
        "--headless",
        "--users",
        "10",
        "--spawn-rate",
        "10",  # Increased spawn rate to reach target faster
        "--run-time",
        "30s",  # Adjusted time to ensure we get exactly 300 requests
        "--only-summary",
        "--json",
    ]

    result = subprocess.run(cmd, capture_output=True, text=True)

    # Save results
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    results_dir = "results"
    os.makedirs(results_dir, exist_ok=True)

    with open(f"{results_dir}/{framework_name}_{timestamp}.json", "w") as f:
        f.write(result.stdout)

    return result.stdout


def main():
    # Clean previous results before starting new benchmarks
    print("Cleaning previous results...")
    clean_previous_results()

    # Start all servers
    servers = []
    frameworks = {
        "fastapi": 8000,
        "flask": 8001,
        "sanic": 8002,
        "quart": 8003,
        "nexios": 8005,
    }

    # Start servers and wait for them to be ready
    for framework, port in frameworks.items():
        print(f"Starting {framework} server...")
        server_file = f"apps/{framework}_app.py"
        server = run_server(server_file, port)
        servers.append(server)

        # Wait for server to be ready
        url = f"http://127.0.0.1:{port}/"
        if wait_for_server(url):
            print(f"{framework} server is ready")
        else:
            print(f"Warning: {framework} server failed to start properly")

    try:
        # Run benchmarks
        results = {}
        for framework in frameworks:
            print(f"\nBenchmarking {framework}...")
            test_file = f"locustfiles/{framework}_test.py"
            results[framework] = run_locust_test(test_file, framework)
            print(f"Completed {framework} benchmark")

    finally:
        # Cleanup: stop all servers
        print("\nStopping servers...")
        for server in servers:
            server.terminate()
            server.wait()

        # Run analysis
        print("\nAnalyzing results...")
        subprocess.run(["uv", "run", "analyze_results.py"])


if __name__ == "__main__":
    main()



import contextvars
import inspect
from functools import wraps
from inspect import Parameter, signature
from typing import Any, Callable, Dict, List, Optional

from nexios.utils.async_helpers import is_async_callable
from nexios.utils.concurrency import run_in_threadpool


class Depend:
    def __init__(self, dependency: Optional[Callable[..., Any]] = None):
        self.dependency = dependency

    def __class_getitem__(cls, item: Any):
        return cls


class Context:
    def __init__(self, request=None, user=None, **kwargs):
        self.request = request
        self.user = user
        for k, v in kwargs.items():
            setattr(self, k, v)


current_context: contextvars.ContextVar[Context] = contextvars.ContextVar(
    "current_context"
)


def inject_dependencies(handler: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to inject dependencies into a route handler while preserving parameter names and passing context if needed."""

    @wraps(handler)
    async def wrapped(*args: List[Any], **kwargs: Dict[str, Any]) -> Any:
        sig = signature(handler)
        bound_args = sig.bind_partial(*args, **kwargs)
        params = list(sig.parameters.values())

        # Pass context if handler accepts it and not already provided
        ctx = None
        if "context" in sig.parameters and "context" not in bound_args.arguments:
            try:
                ctx = current_context.get()
            except LookupError:
                ctx = None
            bound_args.arguments["context"] = ctx

        # --- Generator/async generator cleanup support ---
        cleanup_callbacks = []
        # Store cleanup in context if possible
        if ctx is not None:
            if not hasattr(ctx, "_dependency_cleanup"):
                ctx._dependency_cleanup = []
            cleanup_callbacks = ctx._dependency_cleanup

        for param in params:
            if (
                param.default != Parameter.empty
                and isinstance(param.default, Depend)
                and param.name not in bound_args.arguments
            ):
                depend = param.default
                dependency_func = depend.dependency

                if dependency_func is None:
                    raise ValueError(
                        f"Dependency for parameter '{param.name}' has no provider"
                    )

                if hasattr(dependency_func, "__wrapped__"):
                    dependency_func = dependency_func.__wrapped__  # type: ignore[attr-defined]

                dep_sig = signature(dependency_func)
                dep_kwargs = {}

                # Pass context to dependencies if they accept it
                if "context" in dep_sig.parameters and "context" not in dep_kwargs:
                    if ctx is None:
                        try:
                            ctx = current_context.get()
                        except LookupError:
                            ctx = None
                    dep_kwargs["context"] = ctx

                for dep_param in dep_sig.parameters.values():
                    if dep_param.name in bound_args.arguments:
                        dep_kwargs[dep_param.name] = bound_args.arguments[
                            dep_param.name
                        ]
                    elif dep_param.default != Parameter.empty and isinstance(
                        dep_param.default, Depend
                    ):
                        nested_dep = dep_param.default.dependency
                        if inspect.iscoroutinefunction(nested_dep):
                            dep_kwargs[dep_param.name] = await nested_dep()
                        else:
                            dep_kwargs[dep_param.name] = nested_dep()  # type: ignore[attr-defined]

                if inspect.isasyncgenfunction(dependency_func):
                    agen = dependency_func(**dep_kwargs)
                    value = await agen.__anext__()
                    bound_args.arguments[param.name] = value
                    cleanup_callbacks.append(lambda agen=agen: agen.aclose())
                elif inspect.isgeneratorfunction(dependency_func):
                    gen = dependency_func(**dep_kwargs)
                    value = next(gen)
                    bound_args.arguments[param.name] = value
                    cleanup_callbacks.append(lambda gen=gen: gen.close())
                elif inspect.iscoroutinefunction(dependency_func):
                    bound_args.arguments[param.name] = await dependency_func(
                        **dep_kwargs
                    )
                else:
                    bound_args.arguments[param.name] = dependency_func(**dep_kwargs)

        try:
            if is_async_callable(handler):
                return await handler(**bound_args.arguments)
            else:
                return await run_in_threadpool(handler, **bound_args.arguments)
        finally:
            # Run cleanup callbacks (sync or async)
            for cleanup in cleanup_callbacks:
                result = cleanup()
                if inspect.isawaitable(result):
                    await result

    return wrapped



import importlib
import os
from pathlib import Path
from typing import List, Optional, TypedDict

from nexios.application import NexiosApp
from nexios.logging import create_logger
from nexios.routing import Routes

logger = create_logger("nexios")


class FileRouterConfig(TypedDict):
    root: str
    exempt_paths: Optional[List[str]]
    exclude_from_schema: Optional[bool]


class FileRouter:
    """
    from nexios import get_application
    from nexios.plugins import FileRouterPlugin

    app = get_application()

    FileRouter(app, config={"root": "./routes"})
    """

    app: NexiosApp
    config: FileRouterConfig

    def __init__(
        self,
        app,
        config: FileRouterConfig = {
            "root": "./routes",
            "exempt_paths": [],
            "exclude_from_schema": False,
        },
    ):
        self.app = app
        self.config = config

        self._setup()

    def _setup(self):
        exempt_paths = set(
            os.path.abspath(path)
            for path in self.config.get("exempt_paths", [])  # type:ignore
        )  # type:ignore
        for root, _, files in os.walk(self.config["root"]):
            if os.path.abspath(root) in exempt_paths:
                continue  # Skip the exempted paths
            for file in files:
                file_path = os.path.join(root, file)
                if not file_path.endswith("route.py"):
                    continue

                for route in self._build_route(file_path):
                    self.app.add_route(route)

    def _get_path(self, route_file_path: str) -> str:
        path = route_file_path.replace("route.py", "")
        segments = [
            "{%s}" % segment.replace("_", "") if segment.startswith("_") else segment
            for segment in path.split("/")
        ]

        return "/".join(segments)

    def restrict_slash(self, s: str) -> bool:
        return s.strip() != "/"

    def _build_route(self, route_file_path: str) -> list[Routes]:
        handlers: list[Routes] = []
        path = self._get_path(route_file_path.replace(self.config["root"], ""))

        # Convert file path to a valid module import path
        module_path = (
            Path(route_file_path).with_suffix("").as_posix().replace("/", ".")
        ).lstrip(".")  # Remove leading dot if present

        module = importlib.import_module(module_path)  # Import dynamically

        for attr_name in dir(module):
            methods = ["get", "post", "patch", "put", "delete"]
            is_route = attr_name in methods or hasattr(
                getattr(module, attr_name), "_is_route"
            )

            if is_route:
                handler_function = getattr(module, attr_name)
                logger.debug(f"Mapped {attr_name} {path}")
                path = getattr(handler_function, "_path", path.replace("\\", "/"))
                if attr_name in ["get", "post", "patch", "put", "delete"]:
                    methods = [attr_name.upper()]
                else:
                    methods = getattr(handler_function, "_allowed_methods", ["GET"])
                handlers.append(
                    Routes(
                        path=path.rstrip("/") if self.restrict_slash(path) else path,
                        handler=handler_function,  # type:ignore
                        methods=methods,
                        name=getattr(handler_function, "_name", ""),
                        summary=getattr(handler_function, "_summary", ""),
                        description=getattr(handler_function, "_description", ""),
                        responses=getattr(handler_function, "_responses", {}),
                        request_model=getattr(handler_function, "_request_model", None),
                        middleware=getattr(handler_function, "_middleware", []),
                        tags=getattr(handler_function, "_tags", []),
                        security=getattr(handler_function, "_security", []),
                        operation_id=getattr(handler_function, "_operation_id", ""),
                        deprecated=getattr(handler_function, "_deprecated", False),
                        parameters=getattr(handler_function, "_parameters", []),
                        exclude_from_schema=getattr(
                            handler_function,
                            "_exclude_from_schema",
                            self.config.get("exclude_from_schema", False),
                        ),  # type: ignore
                    )
                )

        return handlers



#!/usr/bin/env python
"""
Nexios CLI - Interactive shell command.
"""

import sys

import click

from nexios.cli.utils import (
    _echo_error,
    _echo_info,
    _echo_warning,
    _load_app_from_path,
    load_config_module,
)


@click.command()
@click.option(
    "--app",
    "app_path",
    # required=True,
    help="App module path in format 'module:app_variable' (e.g., 'myapp.main:app').",
)
@click.option(
    "--config",
    "config_path",
    help="Path to a Python config file that sets up the app instance.",
)
@click.option(
    "--ipython",
    is_flag=True,
    help="Force use of IPython shell (default: auto-detect)",
)
def shell(app_path: str, config_path: str = None, ipython: bool = False):
    """
    Start an interactive shell with the Nexios app context loaded.

    This provides an interactive environment where you can:
    - Access your app instance as 'app'
    - Test routes and handlers
    - Inspect app configuration
    - Debug and experiment with your application

    Examples:
      nexios shell --app myapp.main:app
      nexios shell --app myapp.main:app --ipython
    """
    try:
        # Load config if provided (will return empty dict if file doesn't exist)
        app, config = load_config_module(config_path)

        # If app_path was provided in config, use it (CLI arg takes precedence)
        if "app_path" in config and not app_path:
            app_path = config["app_path"]

        # Load app instance using the provided path
        app = _load_app_from_path(app_path, config_path)
        if app is None:
            _echo_error(
                "Could not load the app instance. Please check your app_path or config."
            )
            sys.exit(1)

        _echo_info(f"Loaded app: {app}")

        # Prepare the shell environment
        shell_vars = {
            "app": app,
            "NexiosApp": type(app),
            "config": getattr(app, "config", {}),
        }

        # Try to import common modules that might be useful
        try:
            from nexios.testing.client import Client

            shell_vars["Client"] = Client
            _echo_info("Test client available as 'Client'")
        except ImportError:
            pass

        try:
            from nexios.http.request import Request
            from nexios.http.response import Response

            shell_vars["Request"] = Request
            shell_vars["Response"] = Response
            _echo_info("Request/Response classes available")
        except ImportError:
            pass

        try:
            from nexios.config import MakeConfig

            shell_vars["MakeConfig"] = MakeConfig
            _echo_info("MakeConfig available for configuration")
        except ImportError:
            pass

        # Try to start IPython if available or requested
        if ipython:
            if not _try_start_ipython_shell(shell_vars):
                _echo_warning("Falling back to regular Python shell")
                _try_start_regular_shell(shell_vars)
        else:
            if not _try_start_regular_shell(shell_vars):
                _echo_info("IPython not found, trying regular shell")
                _try_start_ipython_shell(shell_vars)

    except Exception as e:
        _echo_error(f"Error starting shell: {e}")
        sys.exit(1)


def _try_start_ipython_shell(shell_vars: dict) -> bool:
    """Try to start IPython shell."""
    try:
        import IPython  # noqa: F401
        from IPython.terminal.embed import InteractiveShellEmbed

        _echo_info("Starting IPython shell...")
        _echo_info(
            "Available variables: app, config, Client, Request, Response, MakeConfig"
        )
        _echo_info("Type 'exit' or press Ctrl+D to exit")

        banner = """
Nexios Interactive Shell
=======================
Available variables:
- app: Your Nexios application instance
- config: Application configuration
- Client: Test client for making requests
- Request: Request class
- Response: Response class
- MakeConfig: Configuration class

Examples:
  # Test a route
  async with Client(app) as client:
      resp = await client.get('/')
      print(resp.status_code)
      
  # Inspect app
  print(app.routes)
  print(app.config)
"""

        shell = InteractiveShellEmbed(banner1=banner)
        shell(local_ns=shell_vars)
        return True

    except ImportError:
        return False


def _try_start_regular_shell(shell_vars: dict) -> bool:
    """Try to start regular Python shell."""
    try:
        import code

        _echo_info("Starting Python shell...")
        _echo_info(
            "Available variables: app, config, Client, Request, Response, MakeConfig"
        )
        _echo_info("Type 'exit()' or press Ctrl+D to exit")

        banner = """
Nexios Interactive Shell
=======================
Available variables:
- app: Your Nexios application instance
- config: Application configuration
- Client: Test client for making requests
- Request: Request class
- Response: Response class
- MakeConfig: Configuration class
"""

        console = code.InteractiveConsole(shell_vars)
        console.interact(banner=banner)
        return True

    except Exception:
        return False


if __name__ == "__main__":
    shell()



import http
import traceback
import typing

from nexios.config import get_config
from nexios.exceptions import NotFoundException
from nexios.http import Request, Response


def generate_html_page(title: str, message: str) -> str:
    """Generates a simple HTML page without using Jinja2."""
    return f"""<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>{title}</title>
    <style>
        body {{
            font-family: Arial, sans-serif;
            text-align: center;
            margin: 50px;
            color: #333;
        }}
        h1 {{
            font-size: 48px;
            color: #d9534f;
        }}
        p {{
            font-size: 18px;
            margin-top: 10px;
        }}
    </style>
</head>
<body>
    <h1>{title}</h1>
    <p>{message}</p>
</body>
</html>"""


async def handle_404_error(
    request: Request,
    response: Response,
    exception: NotFoundException,
) -> typing.Any:
    """
    Handles 404 errors dynamically, supporting JSON, HTML, and plain text responses.

    Args:
        request: The request object (not used here but kept for flexibility).
        response: A callable that returns a response.
        exception: The NotFoundException instance.
        settings: A dictionary of settings to determine the response format.

    Returns:
        A response based on the settings.
    """
    settings = get_config()

    debug = settings.debug or False
    not_found_config = settings.not_fouund

    if not_found_config:
        return_json = not_found_config.return_json
        custom_message = not_found_config.custom_message
        show_traceback = not_found_config.show_traceback
        use_html = not_found_config.use_html
    else:
        return_json = True
        custom_message = "The page you are looking for does not exist."
        show_traceback = False
        use_html = True

    error_message = exception.detail if debug else custom_message

    if show_traceback and debug:
        traceback_info = traceback.format_exc()
    else:
        traceback_info = None

    if return_json:
        error_details: typing.Dict[str, typing.Any] = {
            "status": 404,
            "error": http.HTTPStatus(404).phrase,
            "message": error_message,
        }
        if traceback_info:
            error_details["traceback"] = traceback_info

        return response.json(error_details, status_code=404)

    if use_html:
        html_content = generate_html_page("404 - Not Found", error_message)
        return response.html(html_content, status_code=404)

    # Plain Text Fallback
    return response.text(
        f"404 - Not Found\n{error_message}",
    )



import gzip
import json
from unittest.mock import MagicMock, Mock, patch

import pytest

from nexios import get_application
from nexios.http import Request, Response
from nexios.middleware.base import BaseMiddleware
from nexios.middleware.common import CommonMiddleware
from nexios.middleware.gzip import GZipMiddleware
from nexios.middleware.security import SecurityMiddleware
from nexios.testing import Client


class TestBaseMiddleware:
    """Test BaseMiddleware functionality"""

    def test_base_middleware_initialization(self):
        """Test BaseMiddleware initialization"""
        middleware = BaseMiddleware()
        assert middleware is not None

    def test_base_middleware_call(self):
        """Test BaseMiddleware __call__ method"""
        middleware = BaseMiddleware()
        request = Mock()
        response = Mock()

        # Should return the response as-is
        async def mock_call_next():
            return response

        result = middleware(request, response, mock_call_next)
        assert result


class TestCommonMiddleware:
    """Test CommonMiddleware functionality"""

    @pytest.fixture
    async def app_with_common_middleware(self):
        """Create app with CommonMiddleware"""
        app = get_application()
        app.add_middleware(CommonMiddleware())
        async with Client(app) as client:
            yield client, app

    def test_common_middleware_initialization(self):
        """Test CommonMiddleware initialization"""
        middleware = CommonMiddleware()
        assert middleware is not None

    def test_common_middleware_with_config(self):
        """Test CommonMiddleware with configuration"""
        config = {
            "trusted_hosts": ["example.com"],
            "allowed_hosts": ["*.example.com"],
            "max_content_length": 1024 * 1024,
        }
        middleware = CommonMiddleware(**config)
        assert middleware is not None

    async def test_common_middleware_request_processing(
        self, app_with_common_middleware
    ):
        """Test CommonMiddleware request processing"""
        client, app = app_with_common_middleware

        @app.get("/test")
        async def test_handler(req: Request, res: Response):
            return res.text("OK")

        response = await client.get("/test")
        assert response.status_code == 200
        assert response.text == "OK"

    async def test_common_middleware_trusted_hosts(self, app_with_common_middleware):
        """Test CommonMiddleware with trusted hosts"""
        client, app = app_with_common_middleware

        @app.get("/test")
        async def test_handler(req: Request, res: Response):
            return res.text("OK")

        # Test with valid host
        response = await client.get("/test", headers={"Host": "localhost"})
        assert response.status_code == 200

    async def test_common_middleware_content_length_limit(
        self, app_with_common_middleware
    ):
        """Test CommonMiddleware content length limit"""
        client, app = app_with_common_middleware

        @app.post("/test")
        async def test_handler(req: Request, res: Response):
            return res.text("OK")

        # Test with large content
        large_data = "x" * 1024 * 1024  # 1MB
        response = await client.post("/test", data=large_data)
        # Should handle large content appropriately
        assert response.status_code in [200, 413]  # 413 if limit exceeded


# class TestGipMiddleware:
#     """Test GzipMiddleware functionality"""

#     @pytest.fixture
#     async def app_with_gzip_middleware(self):
#         """Create app with GzipMiddleware"""
#         app = get_application()
#         app.add_middleware(GZipMiddleware())
#         async with Client(app) as client:
#             yield client, app

#     def test_gzip_middleware_initialization(self):
#         """Test GzipMiddleware initialization"""
#         middleware = GZipMiddleware()
#         assert middleware is not None

#     def test_gzip_middleware_with_config(self):
#         """Test GzipMiddleware with configuration"""
#         config = {
#             "compress_level": 6,
#             "min_size": 500,
#             "content_types": ["text/plain", "application/json"],
#         }
#         middleware = GZipMiddleware(**config)
#         assert middleware is not None

#     async def test_gzip_middleware_compression(self, app_with_gzip_middleware):
#         """Test GzipMiddleware compression"""
#         client, app = app_with_gzip_middleware

#         @app.get("/test")
#         async def test_handler(req: Request, res: Response):
#             return res.text("This is a test response that should be compressed")

#         response = await client.get("/test", headers={"Accept-Encoding": "gzip"})
#         assert response.status_code == 200

#         # Check if response is compressed
#         content_encoding = response.headers.get("content-encoding")
#         if content_encoding == "gzip":
#             # Verify it's actually gzipped
#             import gzip
#             decompressed = gzip.decompress(response.content).decode()
#             assert "This is a test response" in decompressed

# TODO: Wrap this to suit nexios

# async def test_gzip_middleware_no_compression_small_content(self, app_with_gzip_middleware):
#     """Test GzipMiddleware with small content (should not compress)"""
#     client, app = app_with_gzip_middleware

#     @app.get("/test")
#     async def test_handler(req: Request, res: Response):
#         return res.text("small")

#     response = await client.get("/test", headers={"Accept-Encoding": "gzip"})
#     assert response.status_code == 200

#     # Small content should not be compressed
#     assert response.headers.get("content-encoding") != "gzip"

# async def test_gzip_middleware_no_accept_encoding(self, app_with_gzip_middleware):
#     """Test GzipMiddleware without Accept-Encoding header"""
#     client, app = app_with_gzip_middleware

#     @app.get("/test")
#     async def test_handler(req: Request, res: Response):
#         return res.text("This is a test response")

#     response = await client.get("/test")  # No Accept-Encoding header
#     assert response.status_code == 200
#     assert response.headers.get("content-encoding") != "gzip"

# async def test_gzip_middleware_json_response(self, app_with_gzip_middleware):
#     """Test GzipMiddleware with JSON response"""
#     client, app = app_with_gzip_middleware

#     @app.get("/test")
#     async def test_handler(req: Request, res: Response):
#         return res.json({"message": "This is a JSON response", "data": [1, 2, 3, 4, 5]})

#     response = await client.get("/test", headers={"Accept-Encoding": "gzip"})
#     assert response.status_code == 200

#     # JSON responses should be compressible
#     content_encoding = response.headers.get("content-encoding")
#     if content_encoding == "gzip":
#         decompressed = gzip.decompress(response.content).decode()
#         data = json.loads(decompressed)
#         assert data["message"] == "This is a JSON response"


class TestSecurityMiddleware:
    """Test SecurityMiddleware functionality"""

    @pytest.fixture
    async def app_with_security_middleware(self):
        """Create app with SecurityMiddleware"""
        app = get_application()
        app.add_middleware(SecurityMiddleware())
        async with Client(app) as client:
            yield client, app

    def test_security_middleware_initialization(self):
        """Test SecurityMiddleware initialization"""
        middleware = SecurityMiddleware()
        assert middleware is not None

    def test_security_middleware_with_config(self):
        """Test SecurityMiddleware with configuration"""
        config = {
            "csp_policy": "default-src 'self'",
            "frame_options": "DENY",
            "content_type_options": "nosniff",
            "xss_protection": "1; mode=block",
        }
        middleware = SecurityMiddleware(**config)
        assert middleware is not None

    async def test_security_middleware_csp_header(self, app_with_security_middleware):
        """Test SecurityMiddleware Content Security Policy header"""
        client, app = app_with_security_middleware

        @app.get("/test")
        async def test_handler(req: Request, res: Response):
            return res.text("OK")

        response = await client.get("/test")
        assert response.status_code == 200

        # CSP header might be present depending on configuration
        headers = response.headers
        # This is a basic check - actual CSP value depends on middleware config

    async def test_security_middleware_hsts_header(self, app_with_security_middleware):
        """Test SecurityMiddleware HSTS header"""
        client, app = app_with_security_middleware

        @app.get("/test")
        async def test_handler(req: Request, res: Response):
            return res.text("OK")

        response = await client.get("/test")
        assert response.status_code == 200

        # HSTS header might be present depending on configuration
        headers = response.headers
        # This is a basic check - actual HSTS value depends on middleware config


class TestMiddlewareIntegration:
    """Integration tests for middleware"""

    @pytest.fixture
    async def app_with_multiple_middleware(self):
        """Create app with multiple middleware"""
        app = get_application()
        app.add_middleware(SecurityMiddleware())
        # app.add_middleware(GZipMiddleware())
        app.add_middleware(CommonMiddleware())
        async with Client(app) as client:
            yield client, app

    async def test_multiple_middleware_integration(self, app_with_multiple_middleware):
        """Test multiple middleware working together"""
        client, app = app_with_multiple_middleware

        @app.get("/test")
        async def test_handler(req: Request, res: Response):
            return res.json({"message": "Test response", "data": [1, 2, 3, 4, 5]})

        response = await client.get("/test", headers={"Accept-Encoding": "gzip"})
        assert response.status_code == 200

        # Should have security headers
        headers = response.headers
        # Should potentially have compression
        # Should handle the request properly

    async def test_middleware_error_handling(self, app_with_multiple_middleware):
        """Test middleware error handling"""
        client, app = app_with_multiple_middleware

        @app.get("/error")
        async def error_handler(req: Request, res: Response):
            raise Exception("Test error")

        # Should handle errors gracefully
        response = await client.get("/error")
        # Should return appropriate error response
        assert response.status_code in [500, 404]  # Depending on error handling



from locust import HttpUser, between, task


class FastAPIUser(HttpUser):
    wait_time = between(1, 3)
    host = "http://127.0.0.1:8000"

    @task(1)
    def get_root(self):
        self.client.get("/")

    @task(1)
    def get_json(self):
        self.client.get("/json")

    @task(1)
    def post_echo(self):
        payload = {
            "message": "Hello World",
            "number": 42,
            "boolean": True,
            "array": [1, 2, 3],
            "object": {"key": "value"},
        }
        self.client.post("/echo", json=payload)

    @task(1)
    def get_delay(self):
        self.client.get("/delay/0.1")



from __future__ import annotations

import inspect
import typing

from typing_extensions import Annotated, Doc

from nexios.http import Request, Response
from nexios.middleware.base import BaseMiddleware

from .base import AuthenticationBackend, BaseUser, UnauthenticatedUser


class AuthenticationMiddleware(BaseMiddleware):
    """
    Middleware responsible for handling user authentication.

    This middleware intercepts incoming HTTP requests, calls the authentication
    backend to verify user credentials, and attaches the authenticated user to
    the request scope.

    Attributes:
        backend (AuthenticationBackend): The authentication backend used to verify users.
    """

    def __init__(
        self,
        backend: Annotated[
            AuthenticationBackend,
            Doc("The authentication backend responsible for verifying users."),
        ],
    ) -> None:
        """
        Initializes the authentication middleware with a specified backend.

        Args:
            backend (AuthenticationBackend): An instance of the authentication backend.
        """
        self.backend = backend

    async def process_request(
        self,
        request: Annotated[
            Request,
            Doc("The HTTP request object, containing authentication credentials."),
        ],
        response: Annotated[
            Response,
            Doc(
                "The HTTP response object, which may be modified during authentication."
            ),
        ],
        call_next: typing.Callable[..., typing.Awaitable[typing.Any]],
    ) -> None:
        """
        Processes an incoming request by authenticating the user.

        This method calls the authentication backend, determines if the user is authenticated,
        and attaches the authenticated user to the request. If authentication fails, the request
        is assigned an `UnauthenticatedUser` instance.

        Args:
            request (Request): The HTTP request object.
            response (Response): The HTTP response object.

        Side Effects:
            - Sets `request.user` to an authenticated user or `UnauthenticatedUser`.
            - Updates `request.scope["user"]` with the user object.

        """
        if not inspect.iscoroutinefunction(self.backend.authenticate):
            user: typing.Tuple[BaseUser, str] = self.backend.authenticate(
                request, response
            )  # type:ignore
        else:
            user: typing.Tuple[BaseUser, str] = await self.backend.authenticate(
                request, response
            )  # type:ignore

        if user is None:  # type:ignore
            request.scope["user"] = UnauthenticatedUser()
            request.scope["auth"] = "no-auth"
            await call_next()

            return

        request.scope["user"] = user[0]
        request.scope["auth"] = user[-1]

        await call_next()



import datetime
from contextlib import asynccontextmanager

from sqlalchemy import Boolean, Column, DateTime, Integer, String
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from nexios import NexiosApp
from nexios.middleware.base import BaseMiddleware

# Database configuration
DATABASE_URL = "sqlite+aiosqlite:///./example_async.db"

# SQLAlchemy setup
engine = create_async_engine(DATABASE_URL, echo=True)
async_session = sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)

# Declarative base class
Base = declarative_base()


# Define models
class Note(Base):
    __tablename__ = "notes"

    id = Column(Integer, primary_key=True, index=True)
    title = Column(String, index=True)
    content = Column(String)
    is_public = Column(Boolean, default=False)
    created_at = Column(DateTime, default=datetime.datetime.utcnow)
    updated_at = Column(
        DateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow
    )


# Database dependency
@asynccontextmanager
async def get_session() -> AsyncSession:
    async with async_session() as session:
        try:
            yield session
        finally:
            await session.close()


# Initialize app
app = NexiosApp()


# Database initialization middleware
class AsyncDatabaseMiddleware(BaseMiddleware):
    async def __call__(self, request, response, next):
        async with engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)
        return await next()


app.add_middleware(AsyncDatabaseMiddleware())


@app.post("/notes")
async def create_note(request, response):
    data = await request.json

    async with get_session() as session:
        note = Note(
            title=data["title"],
            content=data["content"],
            is_public=data.get("is_public", False),
        )
        session.add(note)
        await session.commit()
        await session.refresh(note)

        return response.json(
            {
                "id": note.id,
                "title": note.title,
                "content": note.content,
                "is_public": note.is_public,
                "created_at": note.created_at.isoformat(),
                "updated_at": note.updated_at.isoformat(),
            },
            status_code=201,
        )


@app.get("/notes")
async def list_notes(request, response):
    show_private = request.query_params.get("show_private", "false").lower() == "true"

    async with get_session() as session:
        query = session.query(Note)
        if not show_private:
            query = query.filter(Note.is_public)

        notes = await query.order_by(Note.created_at.desc()).all()

        return response.json(
            [
                {
                    "id": note.id,
                    "title": note.title,
                    "content": note.content,
                    "is_public": note.is_public,
                    "created_at": note.created_at.isoformat(),
                    "updated_at": note.updated_at.isoformat(),
                }
                for note in notes
            ]
        )


@app.get("/notes/{note_id}")
async def get_note(request, response):
    note_id = request.path_params["note_id"]

    async with get_session() as session:
        note = await session.get(Note, note_id)

        if not note:
            return response.json({"error": "Note not found"}, status_code=404)

        if not note.is_public:
            return response.json({"error": "Note is private"}, status_code=403)

        return response.json(
            {
                "id": note.id,
                "title": note.title,
                "content": note.content,
                "is_public": note.is_public,
                "created_at": note.created_at.isoformat(),
                "updated_at": note.updated_at.isoformat(),
            }
        )


@app.put("/notes/{note_id}")
async def update_note(request, response):
    note_id = request.path_params["note_id"]
    data = await request.json

    async with get_session() as session:
        note = await session.get(Note, note_id)

        if not note:
            return response.json({"error": "Note not found"}, status_code=404)

        # Update fields
        if "title" in data:
            note.title = data["title"]
        if "content" in data:
            note.content = data["content"]
        if "is_public" in data:
            note.is_public = data["is_public"]

        await session.commit()
        await session.refresh(note)

        return response.json(
            {
                "id": note.id,
                "title": note.title,
                "content": note.content,
                "is_public": note.is_public,
                "created_at": note.created_at.isoformat(),
                "updated_at": note.updated_at.isoformat(),
            }
        )


@app.delete("/notes/{note_id}")
async def delete_note(request, response):
    note_id = request.path_params["note_id"]

    async with get_session() as session:
        note = await session.get(Note, note_id)

        if not note:
            return response.json({"error": "Note not found"}, status_code=404)

        await session.delete(note)
        await session.commit()

        return response.json({"message": "Note deleted successfully", "id": note_id})



import asyncio
from pathlib import Path

import pytest

from nexios import NexiosApp
from nexios.static import StaticFiles
from nexios.testing.client import Client


@pytest.mark.asyncio
async def test_static_file_serving():
    app = NexiosApp()
    static_dir = Path(__file__).parent / "static"
    app.register(StaticFiles(directory=static_dir), "/static")

    async with Client(app) as client:
        # Test existing file
        resp = await client.get("/static/example.txt")
        assert resp.status_code == 200
        assert b"welcome to nexios" in resp.content

        # Test missing file
        resp = await client.get("/static/doesnotexist.txt")
        assert resp.status_code == 404



import uvicorn
from fastapi import FastAPI, Request

app = FastAPI()


@app.get("/")
async def root():
    return {"message": "Hello World"}


@app.get("/json")
async def json_endpoint():
    return {
        "string": "Hello World",
        "number": 42,
        "boolean": True,
        "array": [1, 2, 3],
        "object": {"key": "value"},
    }


@app.post("/echo")
async def echo(request: Request):
    data = await request.json()
    return data


@app.get("/delay/{seconds}")
async def delay(seconds: float):
    import asyncio

    await asyncio.sleep(seconds)
    return {"message": f"Delayed for {seconds} seconds"}


if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=8000, log_level="error")



from typing import Tuple

import pytest

from nexios import NexiosApp, get_application
from nexios.exceptions import HTTPException, NotFoundException
from nexios.http import Request, Response
from nexios.testing import Client


@pytest.fixture
async def async_client():
    app = get_application()  # Fresh app instance for each test
    async with Client(app, log_requests=True) as c:
        yield c, app


async def test_default_404_handler(async_client: Tuple[Client, NexiosApp]):
    client, app = async_client

    @app.get("/existing-route")
    async def existing_route(req: Request, res: Response):
        return res.text("OK")

    # Test non-existent route
    response = await client.get("/non-existent-route")
    assert response.status_code == 404
    assert "Not Found" in response.text


async def test_custom_404_handler(async_client: Tuple[Client, NexiosApp]):
    client, app = async_client

    async def custom_404_handler(req: Request, res: Response, exc: NotFoundException):
        return res.json({"error": "Custom not found"}, status_code=404)

    app.add_exception_handler(NotFoundException, custom_404_handler)

    response = await client.get("/non-existent-route")
    assert response.status_code == 404
    assert response.json() == {"error": "Custom not found"}


async def test_http_exception_handling(async_client: Tuple[Client, NexiosApp]):
    client, app = async_client

    @app.get("/test-http-exception")
    async def test_route(req: Request, res: Response):
        raise HTTPException(status_code=403, detail="Access denied")

    response = await client.get("/test-http-exception")
    assert response.status_code == 403
    assert response.json() == "Access denied"


async def test_custom_exception_handler(async_client: Tuple[Client, NexiosApp]):
    client, app = async_client

    class CustomException(Exception):
        pass

    @app.get("/test-custom-exception")
    async def test_route(req: Request, res: Response):
        raise CustomException("Something went wrong")

    async def handle_custom_exception(
        req: Request, res: Response, exc: CustomException
    ):
        return res.json({"error": str(exc)}, status_code=400)

    app.add_exception_handler(CustomException, handle_custom_exception)

    response = await client.get("/test-custom-exception")
    assert response.status_code == 400
    assert response.json() == {"error": "Something went wrong"}


async def test_status_code_exception_handler(async_client: Tuple[Client, NexiosApp]):
    client, app = async_client

    @app.get("/test-status-code")
    async def test_route(req: Request, res: Response):
        raise HTTPException(status_code=418, detail="I'm a teapot")

    async def handle_teapot(req: Request, res: Response, exc: HTTPException):
        return res.json({"message": "This is a teapot"}, status_code=418)

    app.add_exception_handler(418, handle_teapot)

    response = await client.get("/test-status-code")
    assert response.status_code == 418
    assert response.json() == {"message": "This is a teapot"}


async def test_exception_handler_ordering(async_client: Tuple[Client, NexiosApp]):
    client, app = async_client

    class SpecificException(HTTPException):
        pass

    class GeneralException(HTTPException):
        pass

    @app.get("/test-specific")
    async def test_specific(req: Request, res: Response):
        raise SpecificException(status_code=400, detail="Specific error")

    @app.get("/test-general")
    async def test_general(req: Request, res: Response):
        raise GeneralException(status_code=400, detail="General error")

    # Register general handler first
    async def general_handler(req: Request, res: Response, exc: HTTPException):
        return res.text("General handler")

    # Register specific handler second
    async def specific_handler(req: Request, res: Response, exc: SpecificException):
        return res.text("Specific handler")

    app.add_exception_handler(HTTPException, general_handler)
    app.add_exception_handler(SpecificException, specific_handler)

    # Specific exception should use specific handler
    response = await client.get("/test-specific")
    assert response.text == "Specific handler"

    # General exception should use general handler
    response = await client.get("/test-general")
    assert response.text == "General handler"


async def test_exception_with_headers(async_client: Tuple[Client, NexiosApp]):
    client, app = async_client

    @app.get("/test-headers")
    async def test_route(req: Request, res: Response):
        raise HTTPException(
            status_code=403, detail="Forbidden", headers={"X-Custom-Header": "value"}
        )

    response = await client.get("/test-headers")
    assert response.status_code == 403
    assert response.headers["x-custom-header"] == "value"


async def test_middleware_exception_handling(async_client: Tuple[Client, NexiosApp]):
    client, app = async_client

    async def error_middleware(req: Request, res: Response, call_next):
        try:
            return await call_next()
        except Exception as exc:
            return res.json({"middleware_error": str(exc)}, status_code=500)

    app.add_middleware(error_middleware)

    @app.get("/test-middleware-error")
    async def test_route(req: Request, res: Response):
        raise ValueError("Error in route")

    response = await client.get("/test-middleware-error")
    assert response.status_code == 500
    assert response.json() == {"middleware_error": "Error in route"}


async def test_combined_exception_handling(async_client: Tuple[Client, NexiosApp]):
    client, app = async_client

    class CustomError(Exception):
        pass

    async def custom_handler(req: Request, res: Response, exc: CustomError):
        return res.json({"custom": True}, status_code=400)

    app.add_exception_handler(CustomError, custom_handler)

    @app.get("/test-combined")
    async def test_route(req: Request, res: Response):
        raise CustomError("Combined test")

    response = await client.get("/test-combined")
    assert response.status_code == 400
    assert response.json() == {"custom": True}



import pytest

from nexios import NexiosApp, get_application
from nexios.http import Request, Response
from nexios.testing import Client

app: NexiosApp = get_application()


@pytest.fixture(autouse=True)
async def async_client():
    async with Client(app, log_requests=True) as c:
        yield c


async def test_middleware_modifies_response(async_client: Client):

    app.router.routes.clear()
    app.http_middleware.clear()

    async def header_middleware(request: Request, response: Response, call_next):
        await call_next()
        response.set_header("X-Middleware-Added", "True")

    app.add_middleware(header_middleware)

    @app.route("/header-test")
    async def test_route(request: Request, response: Response):
        return response.text("OK")

    # Execute test
    response = await async_client.get("/header-test")
    assert response.status_code == 200
    assert "X-Middleware-Added" in response.headers
    assert response.headers["X-Middleware-Added"] == "True"


async def test_middleware_modifies_request(async_client: Client):

    app.router.routes.clear()
    app.http_middleware.clear()

    async def modify_request(request: Request, response: Response, call_next):
        request.state.text = "modified"
        return await call_next()

    app.add_middleware(modify_request)

    @app.route("/request-test")
    async def request_test(request: Request, response: Response):
        return response.text(request.state.text)

    response = await async_client.get("/request-test")
    assert response.text == "modified"


async def test_middleware_order(async_client: Client):

    app.router.routes.clear()
    app.http_middleware.clear()

    async def first_middleware(equest: Request, response: Response, call_next):
        await call_next()
        response.set_header("X-Order-1", "First")
        return response

    async def second_middleware(equest: Request, response: Response, call_next):
        await call_next()
        response.set_header("X-Order-2", "Second")
        return response

    app.add_middleware(first_middleware)
    app.add_middleware(second_middleware)

    @app.route("/order-test")
    async def order_test(request: Request, response: Response):
        return response.text("OK")

    response = await async_client.get("/order-test")
    assert response.headers["X-Order-1"] == "First"
    assert response.headers["X-Order-2"] == "Second"


async def test_middleware_short_circuit(async_client: Client):

    app.router.routes.clear()
    app.http_middleware.clear()

    async def blocking_middleware(request: Request, response: Response, call_next):
        if request.headers.get("Authorization") != "Valid":
            return response.text("Unauthorized", status_code=403)
        return await call_next()

    app.add_middleware(blocking_middleware)

    @app.route("/protected")
    async def protected_route(request: Request, response: Response):
        return response.text("Secret")

    # Test blocked request
    response = await async_client.get("/protected")
    assert response.status_code == 403
    assert response.text == "Unauthorized"

    # Test allowed request
    response = await async_client.get("/protected", headers={"Authorization": "Valid"})
    assert response.status_code == 200
    assert response.text == "Secret"


async def test_middleware_exception_handling(async_client: Client):

    app.router.routes.clear()
    app.http_middleware.clear()

    async def exception_middleware(request: Request, response: Response, call_next):
        try:
            return await call_next()
        except Exception:
            return response.text("Handled Error", status_code=500)

    app.add_middleware(exception_middleware)

    @app.route("/error-route")
    async def error_route(request: Request, response: Response):
        raise ValueError("Something went wrong")

    response = await async_client.get("/error-route")
    assert response.status_code == 500
    assert response.text == "Handled Error"



from nexios.file_router import mark_as_route
from nexios.http import Request, Response


@mark_as_route(
    path="/homelander",
    methods=["post", "delete", "patch", "put"],
    summary="Home Lander",
)
async def get(req: Request, res: Response):
    pass



import secrets
import typing
import warnings
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Iterable, Union

from nexios.config import MakeConfig, get_config


class BaseSessionInterface:
    modified = False

    accessed = False

    deleted = False

    _session_cache: Dict[str, Any] = {}

    def __init__(self, session_key: str) -> None:
        config = get_config()
        self.session_key = session_key
        if not config.secret_key:
            print("secret_key", config.secret_key)
            warnings.warn(
                "`secret_key` is not set, `secret_key`  is required to use session",
                RuntimeWarning,
            )
            return
        self.config = config
        self.session_config: MakeConfig = config.session

    def __getitem__(self, key: str) -> Any:
        self.accessed = True
        return self._session_cache[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.modified = True
        self.accessed = True
        self._session_cache[key] = value

    def __delitem__(self, key: str) -> None:
        self.modified = True
        self.deleted = True
        del self._session_cache[key]

    def __iter__(self) -> Iterable[str]:
        self.accessed = True
        return iter(self._session_cache)

    def __len__(self) -> int:
        self.accessed = True
        return len(self._session_cache)

    def __contains__(self, key: str) -> bool:
        self.accessed = True
        return key in self._session_cache

    def set_session(self, key: str, value: str):
        self.modified = True
        self.accessed = True

        self._session_cache[key] = value

    def get_session(self, key: str):
        self.accessed = True

        return self._session_cache.get(key, None)

    def get_all(self):
        self.accessed = True
        return self._session_cache.items()

    def delete_session(self, key: str):
        self.modified = True
        self.deleted = True
        if key in self._session_cache:
            del self._session_cache[key]

    def keys(self):
        return self._session_cache.keys()

    def values(self):
        return self._session_cache.values()

    def is_empty(self):
        return self._session_cache.items().__len__() == 0

    async def save(self):
        raise NotImplementedError

    def get_cookie_name(self) -> str:
        """The name of the session cookie. Uses``app.config.SESSION_COOKIE_NAME``."""
        if not self.session_config:
            return "session_id"
        return self.session_config.session_cookie_name or "session_id"

    def get_cookie_domain(self) -> typing.Optional[str]:
        """Returns the domain for which the cookie is valid. Uses `config.SESSION_COOKIE_DOMAIN`."""
        if not self.session_config:
            return None
        return self.session_config.session_cookie_domain

    def get_cookie_path(self) -> Union[str, None]:
        """Returns the path for which the cookie is valid. Uses `config.SESSION_COOKIE_PATH`."""
        if not self.session_config:
            return None
        return self.session_config.session_cookie_path

    def get_cookie_httponly(self) -> typing.Optional[bool]:
        """Returns whether the session cookie should be HTTPOnly. Uses `session_config.session_cookie_httponly`."""
        if not self.session_config:
            return None
        return self.session_config.session_cookie_httponly

    def get_cookie_secure(self) -> typing.Optional[bool]:
        """Returns whether the session cookie should be secure. Uses `session_config.session_cookie_secure`."""
        if not self.session_config:
            return None
        return self.session_config.session_cookie_secure

    def get_cookie_samesite(self) -> typing.Optional[str]:
        """Returns the SameSite attribute for the cookie. Uses `session_config.session_cookie_samesite`."""
        if not self.session_config:
            return None
        return self.session_config.session_cookie_samesite

    def get_cookie_partitioned(self) -> typing.Optional[bool]:
        """Returns whether the cookie should be partitioned. Uses `session_config.session_cookie_partitioned`."""
        if not self.session_config:
            return None
        return self.session_config.session_cookie_partitioned

    def get_expiration_time(self) -> typing.Optional[datetime]:
        """Returns the expiration time for the session if one is set."""
        if not self.session_config:
            # No config - use default 7 day expiration
            if not hasattr(self, "_expiration_time"):
                self._expiration_time = datetime.now(timezone.utc) + timedelta(days=7)
            return self._expiration_time

        if not self.session_config.session_permanent:
            # Non-permanent session - calculate expiration if not set
            if not hasattr(self, "_expiration_time"):
                expiration_seconds = (
                    self.session_config.session_expiration_time or 86400
                )
                self._expiration_time = datetime.now(timezone.utc) + timedelta(
                    seconds=expiration_seconds
                )
            return self._expiration_time

        return None

    @property
    def should_set_cookie(self) -> bool:
        """Determines if the cookie should be set. Depends on `config.SESSION_REFRESH_EACH_REQUEST`."""

        if not self.session_config:
            return self.modified
        return self.modified or (
            self.session_config.session_permanent
            and self.session_config.session_refresh_each_request
        )

    def has_expired(self) -> bool:
        """Returns True if the session has expired."""
        expiration_time = self.get_expiration_time()
        if expiration_time and datetime.now(timezone.utc) > expiration_time:  # type: ignore
            return True
        return False

    def get_session_key(self) -> str:
        """Returns the session key."""
        if self.session_key:
            return self.session_key
        return secrets.token_hex(32)

    def clear(self):
        self._session_cache = {}

    def get(self, key: str):
        return self._session_cache.get(key)

    def set_expiration_time(self, expiration: datetime) -> None:
        """Set the expiration time for the session."""
        self._expiration_time = expiration

    def __str__(self) -> str:
        return f"<Sesion {self._session_cache} >"

    def __del__(self) -> None:
        self.clear()
        self.session_key = None
        self.config = None



import sys
from pathlib import Path

from nexios.file_router.html import configure_templates

BASE_DIR = Path(__file__).parent.parent.parent
sys.path.append(str(BASE_DIR))

from nexios import get_application
from nexios.file_router import FileRouter

app = get_application()

FileRouter(app, config={"root": "./routes", "exempt_paths": ["./routes/posts"]})

configure_templates(template_dir="templates")



import mimetypes
import os
import uuid
from pathlib import Path

import aiofiles

from nexios import NexiosApp
from nexios.responses import FileResponse, StreamingResponse

app = NexiosApp()

# Configure upload directory
UPLOAD_DIR = Path("uploads")
UPLOAD_DIR.mkdir(exist_ok=True)


def get_safe_filename(filename: str) -> str:
    """Generate a safe filename with UUID to prevent conflicts"""
    ext = Path(filename).suffix
    return f"{uuid.uuid4()}{ext}"


async def save_upload_file(file, filename: str) -> Path:
    """Save an uploaded file to disk"""
    filepath = UPLOAD_DIR / filename
    async with aiofiles.open(filepath, "wb") as f:
        while chunk := await file.read(8192):
            await f.write(chunk)
    return filepath


@app.post("/upload")
async def upload_file(request, response):
    # Get uploaded files
    files = await request.files

    if not files:
        return response.json({"error": "No files uploaded"}, status_code=400)

    uploaded_files = []
    for file_field, file in files.items():
        # Generate safe filename
        safe_filename = get_safe_filename(file.filename)

        # Save file
        filepath = await save_upload_file(file, safe_filename)

        uploaded_files.append(
            {
                "original_name": file.filename,
                "saved_name": safe_filename,
                "content_type": file.content_type,
                "size": os.path.getsize(filepath),
            }
        )

    return response.json(
        {
            "message": f"Successfully uploaded {len(uploaded_files)} files",
            "files": uploaded_files,
        }
    )


@app.get("/files")
async def list_files(request, response):
    files = []
    for filepath in UPLOAD_DIR.glob("*"):
        if filepath.is_file():
            files.append(
                {
                    "name": filepath.name,
                    "size": filepath.stat().st_size,
                    "content_type": mimetypes.guess_type(filepath)[0],
                    "uploaded_at": datetime.fromtimestamp(
                        filepath.stat().st_mtime
                    ).isoformat(),
                }
            )

    return response.json({"files": files, "total": len(files)})


@app.get("/files/{filename}")
async def download_file(request, response):
    filename = request.path_params["filename"]
    filepath = UPLOAD_DIR / filename

    if not filepath.exists():
        return response.json({"error": "File not found"}, status_code=404)

    return FileResponse(
        filepath, filename=filename, content_type=mimetypes.guess_type(filepath)[0]
    )


@app.get("/stream/{filename}")
async def stream_file(request, response):
    filename = request.path_params["filename"]
    filepath = UPLOAD_DIR / filename

    if not filepath.exists():
        return response.json({"error": "File not found"}, status_code=404)

    async def file_streamer():
        async with aiofiles.open(filepath, "rb") as f:
            while chunk := await f.read(8192):
                yield chunk

    return StreamingResponse(
        file_streamer(), media_type=mimetypes.guess_type(filepath)[0]
    )


@app.delete("/files/{filename}")
async def delete_file(request, response):
    filename = request.path_params["filename"]
    filepath = UPLOAD_DIR / filename

    if not filepath.exists():
        return response.json({"error": "File not found"}, status_code=404)

    try:
        filepath.unlink()
        return response.json({"message": f"Successfully deleted {filename}"})
    except Exception as e:
        return response.json(
            {"error": f"Failed to delete file: {str(e)}"}, status_code=500
        )


# Serve a simple upload form
@app.get("/")
async def upload_form(request, response):
    html = """
    <!DOCTYPE html>
    <html>
        <head>
            <title>File Upload Example</title>
            <style>
                body { font-family: Arial, sans-serif; max-width: 800px; margin: 0 auto; padding: 20px; }
                .upload-form { border: 2px dashed #ccc; padding: 20px; margin: 20px 0; }
                .file-list { margin-top: 20px; }
                .file-item { border: 1px solid #eee; padding: 10px; margin: 5px 0; }
            </style>
        </head>
        <body>
            <h1>File Upload Example</h1>
            
            <div class="upload-form">
                <h2>Upload Files</h2>
                <form action="/upload" method="post" enctype="multipart/form-data">
                    <input type="file" name="files" multiple>
                    <button type="submit">Upload</button>
                </form>
            </div>
            
            <div class="file-list">
                <h2>Uploaded Files</h2>
                <div id="files"></div>
            </div>
            
            <script>
                // Fetch and display uploaded files
                async function loadFiles() {
                    const response = await fetch('/files');
                    const data = await response.json();
                    
                    const filesDiv = document.getElementById('files');
                    filesDiv.innerHTML = data.files.map(file => `
                        <div class="file-item">
                            <strong>${file.name}</strong>
                            <br>
                            Size: ${(file.size / 1024).toFixed(2)} KB
                            <br>
                            Type: ${file.content_type || 'unknown'}
                            <br>
                            <a href="/files/${file.name}" target="_blank">Download</a>
                            |
                            <a href="/stream/${file.name}" target="_blank">Stream</a>
                            |
                            <button onclick="deleteFile('${file.name}')">Delete</button>
                        </div>
                    `).join('');
                }
                
                async function deleteFile(filename) {
                    if (!confirm('Are you sure you want to delete this file?')) return;
                    
                    await fetch(`/files/${filename}`, { method: 'DELETE' });
                    loadFiles();
                }
                
                // Load files on page load
                loadFiles();
            </script>
        </body>
    </html>
    """
    return response.html(html)



# test_events.py
import asyncio
from unittest.mock import Mock, call

import pytest

from nexios import get_application
from nexios.events import (
    EventEmitter,
    EventPhase,
    EventPriority,
)


# Fixture for app with event emitter, ensuring clean state
@pytest.fixture
def app_with_emitter():
    app = get_application()
    app.emitter = EventEmitter()
    yield app
    app.emitter.remove_all_events()


# Fixture for a mock object to track listener calls
@pytest.fixture
def listener_mock():
    return Mock()


# Test basic event registration and triggering
@pytest.mark.asyncio
async def test_app_event_registration_and_trigger(app_with_emitter, listener_mock):
    app = app_with_emitter

    async def startup_handler():
        listener_mock("started")

    # Register listener explicitly
    app.emitter.event("app:startup").listen(startup_handler)

    # Trigger event and ensure completion
    stats = app.emitter.emit("app:startup")

    # Allow async tasks to complete
    await asyncio.sleep(0)  # Yield to event loop

    assert stats["listeners_executed"] == 1
    assert listener_mock.call_args_list == [call("started")]


# Test one-time listeners
@pytest.mark.asyncio
async def test_app_one_time_listener(app_with_emitter, listener_mock):
    app = app_with_emitter

    async def request_handler():
        listener_mock("handled")

    # Register one-time listener
    app.emitter.event("app:request").once(request_handler)

    # Trigger twice
    stats1 = app.emitter.emit("app:request")
    stats2 = app.emitter.emit("app:request")

    # Allow async tasks to complete
    await asyncio.sleep(0)

    assert stats1["listeners_executed"] == 1
    assert stats2["listeners_executed"] == 0  # Listener removed after first trigger
    assert listener_mock.call_args_list == [call("handled")]


# Test event priorities
@pytest.mark.asyncio
async def test_app_event_priorities(app_with_emitter, listener_mock):
    app = app_with_emitter

    async def high_priority():
        listener_mock("high")

    async def low_priority():
        listener_mock("low")

    # Register listeners with different priorities
    app.emitter.event("app:process").listen(high_priority, priority=EventPriority.HIGH)
    app.emitter.event("app:process").listen(low_priority, priority=EventPriority.LOW)

    # Trigger event
    stats = app.emitter.emit("app:process")

    # Allow async tasks to complete
    await asyncio.sleep(0)

    assert stats["listeners_executed"] == 2
    assert listener_mock.call_args_list == [call("high"), call("low")]


# Test event propagation (parent-child)
@pytest.mark.asyncio
async def test_app_event_propagation(app_with_emitter, listener_mock):
    app = app_with_emitter
    parent_event = app.emitter.event("app:parent")
    child_event = app.emitter.event("app:child")
    child_event.parent = parent_event

    async def parent_handler(*args, phase=EventPhase.AT_TARGET):
        listener_mock(f"parent-{phase.name}")

    async def child_handler(*args, phase=EventPhase.AT_TARGET):
        listener_mock(f"child-{phase.name}")

    # Register listeners
    parent_event.listen(parent_handler)
    child_event.listen(child_handler)

    # Trigger child event
    stats = app.emitter.emit("app:child")

    # Allow async tasks to complete
    await asyncio.sleep(0)

    assert stats["listeners_executed"] == 1  # Only child listeners at target phase



import asyncio
import typing
import warnings
from typing import Any, Callable, Dict, List, Optional

from typing_extensions import Annotated, Doc

from nexios._internals._route_builder import RouteBuilder
from nexios.dependencies import inject_dependencies
from nexios.types import (
    ASGIApp,
    Receive,
    Scope,
    Send,
    WsHandlerType,
    WsMiddlewareType,
)
from nexios.websockets import WebSocket
from nexios.websockets.errors import WebSocketErrorMiddleware

from ._utils import get_route_path
from .base import BaseRouter


class WebsocketRoutes:
    def __init__(
        self,
        path: str,
        handler: WsHandlerType,
        middleware: typing.List[WsMiddlewareType] = [],
    ):
        assert callable(handler), "Route handler must be callable"
        assert asyncio.iscoroutinefunction(handler), "Route handler must be async"
        self.raw_path = path
        self.handler: WsHandlerType = inject_dependencies(handler)
        self.middleware = middleware
        self.route_info = RouteBuilder.create_pattern(path)
        self.pattern = self.route_info.pattern
        self.param_names = self.route_info.param_names
        self.route_type = self.route_info.route_type
        self.router_middleware = None

    def match(self, path: str) -> typing.Tuple[Any, Any]:
        """
        Match a path against this route's pattern and return captured parameters.

        Args:
            path: The URL path to match.

        Returns:
            Optional[Dict[str, Any]]: A dictionary of captured parameters if the path matches,
            otherwise None.
        """
        match = self.pattern.match(path)
        if match:
            matched_params = match.groupdict()
            for key, value in matched_params.items():
                matched_params[key] = self.route_info.convertor[  # type:ignore
                    key
                ].convert(value)
            return match, matched_params
        return None, None

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handles the WebSocket connection by calling the route's handler.

        Args:
            websocket: The WebSocket connection.
            params: The extracted route parameters.
        """
        websocket_session = WebSocket(scope, receive=receive, send=send)
        await self.handler(websocket_session)

    def __repr__(self) -> str:
        return f"<WSRoute {self.raw_path}>"


class WSRouter(BaseRouter):
    def __init__(
        self,
        prefix: Optional[str] = None,
        middleware: Optional[List[Any]] = [],
        routes: Optional[List[WebsocketRoutes]] = [],
    ):
        self.prefix = prefix or ""
        self.routes: List[WebsocketRoutes] = routes or []
        self.middleware: List[Callable[[ASGIApp], ASGIApp]] = []
        self.sub_routers: Dict[str, ASGIApp] = {}
        if self.prefix and not self.prefix.startswith("/"):
            warnings.warn("WSRouter prefix should start with '/'")
            self.prefix = f"/{self.prefix}"

    def add_ws_route(
        self,
        route: Annotated[
            WebsocketRoutes,
            Doc("An instance of the Routes class representing a WebSocket route."),
        ],
    ) -> None:
        """
        Adds a WebSocket route to the application.

        This method registers a WebSocket route, allowing the application to handle WebSocket connections.

        Args:
            route (Routes): The WebSocket route configuration.

        Returns:
            None

        Example:
            ```python
            route = Routes("/ws/chat", chat_handler)
            app.add_ws_route(route)
            ```
        """
        self.routes.append(route)

    def add_ws_middleware(self, middleware: type[ASGIApp]) -> None:  # type: ignore[override]
        """Add middleware to the WebSocket router"""
        self.middleware.insert(0, middleware)  # type: ignore

    def ws_route(
        self,
        path: Annotated[
            str, Doc("The WebSocket route path. Must be a valid URL pattern.")
        ],
        handler: Annotated[
            Optional[WsHandlerType],
            Doc("The WebSocket handler function. Must be an async function."),
        ] = None,
        middleware: Annotated[
            List[WsMiddlewareType],
            Doc("List of middleware to be executes before the router handler"),
        ] = [],
    ) -> Any:
        """
        Registers a WebSocket route.

        This decorator is used to define WebSocket routes in the application, allowing handlers
        to manage WebSocket connections. When a WebSocket client connects to the given path,
        the specified handler function will be executed.

        Returns:
            Callable: The original WebSocket handler function.

        Example:
            ```python

            @app.ws_route("/ws/chat")
            async def chat_handler(websocket):
                await websocket.accept()
                while True:
                    message = await websocket.receive_text()
                    await websocket.send_text(f"Echo: {message}")
            ```
        """
        if handler:
            return self.add_ws_route(
                WebsocketRoutes(path, handler, middleware=middleware)
            )

        def decorator(handler: WsHandlerType) -> WsHandlerType:
            self.add_ws_route(WebsocketRoutes(path, handler, middleware=middleware))
            return handler

        return decorator

    def build_middleware_stack(  # type:ignore
        self, scope: Scope, receive: Receive, send: Send
    ) -> ASGIApp:  # type:ignore
        app = self.app
        for mdw in reversed(self.middleware):
            app = mdw(app)  # type:ignore[assignment]
        return app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "websocket":
            return
        app = self.build_middleware_stack(scope, receive, send)
        app = WebSocketErrorMiddleware(app)
        await app(scope, receive, send)

    async def app(self, scope: Scope, receive: Receive, send: Send) -> None:
        url = get_route_path(scope)
        for mount_path, sub_app in self.sub_routers.items():
            if url.startswith(mount_path):
                scope["path"] = url[len(mount_path) :]
                await sub_app(scope, receive, send)
                return
        for route in self.routes:
            match, params = route.match(url)
            if match:
                scope["route_params"] = params
                await route.handle(scope, receive, send)
                return
        await send({"type": "websocket.close", "code": 404})

    def wrap_asgi(
        self,
        middleware_cls: Annotated[
            Callable[[ASGIApp], Any],
            Doc(
                "An ASGI middleware class or callable that takes an app as its first argument and returns an ASGI app"
            ),
        ],
    ) -> None:
        """
        Wraps the entire application with an ASGI middleware.

        This method allows adding middleware at the ASGI level, which intercepts all requests
        (HTTP, WebSocket, and Lifespan) before they reach the application.

        Args:
            middleware_cls: An ASGI middleware class or callable that follows the ASGI interface
            *args: Additional positional arguments to pass to the middleware
            **kwargs: Additional keyword arguments to pass to the middleware

        Returns:
            NexiosApp: The application instance for method chaining


        """
        self.app = middleware_cls(self.app)

    def mount_router(  # type:ignore
        self, app: "WSRouter", path: typing.Optional[str] = None
    ) -> None:  # type:ignore
        """
        Mount an ASGI application (e.g., another Router) under a specific path prefix.

        Args:
            path: The path prefix under which the app will be mounted.
            app: The ASGI application (e.g., another Router) to mount.
        """

        if not path:
            path = app.prefix
        path = path.rstrip("/")

        if path == "":
            self.sub_routers[path] = app
            return
        if not path.startswith("/"):
            path = f"/{path}"

        self.sub_routers[path] = app

    def __repr__(self) -> str:
        return f"<WSRouter prefix='{self.prefix}' routes={len(self.routes)}>"



# test_cors_with_app.py
import pytest

from nexios import get_application
from nexios.config import MakeConfig
from nexios.middleware.cors import CORSMiddleware
from nexios.testing import Client


@pytest.fixture
async def cors_app():
    # Create app with CORS configuration
    config = MakeConfig(
        {
            "cors": {
                "allow_origins": ["http://example.com", "https://example.org"],
                "allow_methods": ["GET", "POST"],
                "allow_headers": ["X-Custom-Header"],
                "allow_credentials": True,
                "expose_headers": ["X-Exposed-Header"],
                "debug": True,
            }
        }
    )
    app = get_application(config)

    # Add test route
    @app.get("/test")
    async def test_route(req, res):
        return res.text("OK")

    # Add CORS middleware
    app.add_middleware(CORSMiddleware())

    return app


@pytest.fixture
async def client(cors_app):
    async with Client(cors_app) as c:
        yield c


async def test_simple_request_allowed_origin(client):
    # Test simple GET request with allowed origin
    response = await client.get("/test", headers={"Origin": "http://example.com"})

    assert response.status_code == 200
    assert response.text == "OK"
    assert response.headers["Access-Control-Allow-Origin"] == "http://example.com"
    assert response.headers["Access-Control-Allow-Credentials"] == "true"
    assert response.headers["Access-Control-Expose-Headers"] == "X-Exposed-Header"


async def test_simple_request_disallowed_origin(client):
    # Test simple GET request with disallowed origin
    response = await client.get("/test", headers={"Origin": "http://disallowed.com"})

    assert response.status_code == 200
    assert response.text == "OK"
    # Should not have CORS headers for disallowed origin
    assert "Access-Control-Allow-Origin" not in response.headers


async def test_preflight_request_success(client):
    # Test successful OPTIONS preflight request
    response = await client.options(
        "/test",
        headers={
            "Origin": "http://example.com",
            "Access-Control-Request-Method": "GET",
            "Access-Control-Request-Headers": "X-Custom-Header",
        },
    )

    assert response.status_code == 201  # Custom status from our middleware
    assert response.headers["Access-Control-Allow-Origin"] == "http://example.com"
    assert response.headers["Access-Control-Allow-Methods"] == "GET"
    assert response.headers["Access-Control-Allow-Headers"] == "x-custom-header"
    assert response.headers["Access-Control-Allow-Credentials"] == "true"
    assert response.headers["Access-Control-Max-Age"] == "600"  # Default value


async def test_preflight_request_disallowed_method(client):
    # Test OPTIONS request with disallowed method
    response = await client.options(
        "/test",
        headers={
            "Origin": "http://example.com",
            "Access-Control-Request-Method": "PUT",
        },
    )

    # assert response.status_code == 400  # Default error status
    assert "PUT" not in response.headers.get("Access-Control-Allow-Methods", "")


async def test_preflight_request_disallowed_header(client):
    # Test OPTIONS request with disallowed header
    response = await client.options(
        "/test",
        headers={
            "Origin": "http://example.com",
            "Access-Control-Request-Method": "GET",
            "Access-Control-Request-Headers": "X-Disallowed-Header",
        },
    )

    assert response.status_code == 400
    assert "X-Disallowed-Header" not in response.headers.get(
        "Access-Control-Allow-Headers", ""
    )


async def test_wildcard_origin():
    # Test app with wildcard origin
    config = MakeConfig({"cors": {"allow_origins": ["*"], "allow_methods": ["*"]}})
    app = get_application(config)

    @app.get("/wildcard")
    async def wildcard_route(req, res):
        return res.text("OK")

    app.add_middleware(CORSMiddleware())

    async with Client(app) as client:
        response = await client.get(
            "/wildcard", headers={"Origin": "http://any-origin.com"}
        )

        assert response.status_code == 200
        assert (
            response.headers["Access-Control-Allow-Origin"] == "http://any-origin.com"
        )


async def test_no_cors_headers_without_origin():
    # Test that CORS headers aren't added when no Origin header is present
    config = MakeConfig(
        {"cors": {"allow_origins": ["http://example.com"], "allow_methods": ["GET"]}}
    )
    app = get_application(config)

    @app.get("/no-origin")
    async def no_origin_route(req, res):
        return res.text("OK")

    app.add_middleware(CORSMiddleware())

    async with Client(app) as client:
        response = await client.get("/no-origin")

        assert response.status_code == 200
        assert "Access-Control-Allow-Origin" not in response.headers



from .base import BaseRouter
from .grouping import Group
from .http import Router, Routes
from .websocket import WebsocketRoutes, WSRouter

__all__ = [
    "Router",
    "Routes",
    "WSRouter",
    "WebsocketRoutes",
    "BaseRouter",
    "Group",
]



import logging
import types
from nexios.logging import create_logger, has_level_handler, LocalQueueHandler

def test_create_logger_returns_logger():
    logger = create_logger("test_logger")
    assert isinstance(logger, logging.Logger)
    assert logger.name == "test_logger"

def test_logger_has_level_handler():
    logger = create_logger("test_logger2")
    assert has_level_handler(logger)

def test_local_queue_handler_prepare():
    handler = LocalQueueHandler(None)
    record = logging.LogRecord("test", logging.INFO, __file__, 1, "msg", (), None)
    assert handler.prepare(record) is record 


from __future__ import annotations

from functools import wraps
from typing import Dict, List, Optional, Type, Union,TYPE_CHECKING
from uuid import uuid4

from pydantic import BaseModel

from nexios.http import Request, Response

from .config import OpenAPIConfig
from .models import (
    Components,
    ExternalDocumentation,
    MediaType,
    Operation,
    Parameter,
    PathItem,
    RequestBody,
)
from .models import Response as OpenAPIResponse
from .models import (
    Schema,
)
if TYPE_CHECKING:
    from nexios import NexiosApp


class APIDocumentation:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(
        self,
        app: Optional["NexiosApp"] = None,  # noqa: F821
        config: Optional[OpenAPIConfig] = None,
        swagger_url: str = "/docs",
        redoc_url: str = "/redoc",
        openapi_url: str = "/openapi.json",
    ):
        self.app = app
        self.config = config or OpenAPIConfig()
        self.swagger_url = swagger_url
        self.redoc_url = redoc_url
        self.openapi_url = openapi_url
        if app:
            self._setup_doc_routes()

    def _setup_doc_routes(self):
        """Set up routes for serving OpenAPI specification"""

        @self.app.get(self.openapi_url, exclude_from_schema=True)  # type:ignore
        async def serve_openapi(request: Request, response: Response):
            openapi_json = self.config.openapi_spec.model_dump(
                by_alias=True, exclude_none=True
            )
            return response.json(openapi_json)

        @self.app.get(self.swagger_url, exclude_from_schema=True)  # type:ignore
        async def swagger_ui(request: Request, response: Response):
            return response.html(self._generate_swagger_ui())

        @self.app.get(self.redoc_url, exclude_from_schema=True)  # type:ignore
        async def redoc_ui(request: Request, response: Response):
            return response.html(self._generate_redoc_ui())

    @classmethod
    def get_instance(cls):
        return cls._instance

    def _generate_redoc_ui(self) -> str:
        """Generate ReDoc UI HTML"""
        return f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>{self.config.openapi_spec.info.title} - API Documentation</title>
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">
            <style>
                body {{
                    margin: 0;
                    padding: 0;
                }}
            </style>
        </head>
        <body>
            <div id="redoc"></div>
            <script src="https://cdn.redoc.ly/redoc/latest/bundles/redoc.standalone.js"></script>
            <script>
                Redoc.init('{self.openapi_url}', {{
                    scrollYOffset: 50
                }}, document.getElementById('redoc'))
            </script>
        </body>
        </html>
        """

    def _generate_swagger_ui(self) -> str:
        """Generate Swagger UI HTML"""
        return f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>{self.config.openapi_spec.info.title} - Docs</title>
            <link rel="stylesheet" href="https://unpkg.com/swagger-ui-dist@4.18.3/swagger-ui.css">
        </head>
        <body>
            <div id="swagger-ui"></div>
            <script src="https://unpkg.com/swagger-ui-dist@4/swagger-ui-bundle.js"></script>
            <script>
                window.onload = function() {{
                    SwaggerUIBundle({{
                        url: '{self.openapi_url}',
                        dom_id: '#swagger-ui',
                        presets: [
                            SwaggerUIBundle.presets.apis,
                            SwaggerUIBundle.SwaggerUIStandalonePreset
                        ],
                        layout: "BaseLayout"
                    }});
                }}
            </script>
        </body>
        </html>
        """

    def document_endpoint(
        self,
        path: str,
        method: str,
        summary: str = "",
        description: Optional[str] = None,
        parameters: Optional[List[Parameter]] = None,
        request_body: Optional[Type[BaseModel]] = None,
        responses: Optional[Union[BaseModel, Dict[int, BaseModel]]] = None,
        tags: Optional[List[str]] = None,
        security: Optional[List[Dict[str, List[str]]]] = None,
        operation_id: Optional[str] = None,
        deprecated: bool = False,
        external_docs: Optional[ExternalDocumentation] = None,
    ):
        """
        Decorator to document API endpoints with OpenAPI specification

        :param path: URL path of the endpoint
        :param method: HTTP method (get, post, put, delete, etc.)
        :param summary: Short summary of the endpoint
        :param description: Detailed description of the endpoint
        :param parameters: List of parameters
        :param request_body: Pydantic model for request body
        :param responses: Response model(s)
        :param tags: Categorization tags for the endpoint
        :param security: Security requirements
        :param operation_id: Unique identifier for the operation
        :param deprecated: Whether the endpoint is deprecated
        :param external_docs: External documentation reference
        """

        def decorator(func):
            # Prepare request body specification
            request_body_spec = None

            if request_body:
                self.add_schema(request_body)
            if request_body:
                request_body_spec = RequestBody(
                    content={
                        "application/json": MediaType(  # type:ignore
                            schema=Schema(  # type:ignore
                                **request_body.model_json_schema()
                            )
                        )
                    }
                )
            else:
                if method.upper() not in ["GET", "DELETE"]:
                    request_body_spec = RequestBody(
                        content={
                            "application/json": MediaType(  # type:ignore
                                schema=Schema(
                                    example={
                                        "example": "This is an example request body"
                                    }
                                )  # type:ignore
                            )
                        }
                    )

            responses_spec = {}
            if responses:
                if isinstance(responses, type) and issubclass(responses, BaseModel):
                    responses_spec["200"] = OpenAPIResponse(
                        description="Successful Response",
                        content={
                            "application/json": MediaType(  # type:ignore
                                schema=Schema(  # type:ignore
                                    **responses.model_json_schema(),
                                )
                            )
                        },
                    )
                # Handle List[Model] case
                elif hasattr(responses, "__origin__") and responses.__origin__ is list:
                    item_model = responses.__args__[0]
                    if issubclass(item_model, BaseModel):
                        responses_spec["200"] = OpenAPIResponse(
                            description="Successful Response",
                            content={
                                "application/json": MediaType(  # type:ignore
                                    schema=Schema(  # type:ignore
                                        type="array",
                                        items=Schema(**item_model.model_json_schema()),
                                    )
                                )
                            },
                        )
                # Multiple responses case
                elif isinstance(responses, dict):
                    for status_code, model in responses.items():
                        # Handle direct model
                        if isinstance(model, type) and issubclass(model, BaseModel):
                            responses_spec[str(status_code)] = OpenAPIResponse(
                                description=f"Response for status code {status_code}",
                                content={
                                    "application/json": MediaType(  # type:ignore
                                        schema=Schema(  # type:ignore
                                            **model.model_json_schema()
                                        )
                                    )
                                },
                            )
                        # Handle List[Model]
                        elif hasattr(model, "__origin__") and model.__origin__ is list:
                            item_model = model.__args__[0]
                            if issubclass(item_model, BaseModel):
                                responses_spec[str(status_code)] = OpenAPIResponse(
                                    description=f"Response for status code {status_code}",
                                    content={
                                        "application/json": MediaType(  # type:ignore
                                            schema=Schema(  # type:ignore
                                                type="array",
                                                items=Schema(
                                                    **item_model.model_json_schema()
                                                ),
                                            )
                                        )
                                    },
                                )
            else:
                # Default response if no responses specified
                responses_spec["200"] = OpenAPIResponse(
                    description="Successful Response",
                    content={
                        "application/json": MediaType(  # type:ignore
                            schema=Schema(
                                example={"example": "This is an example response"},
                                type="object",
                            )  # type:ignore
                        )
                    },
                )

            if not responses_spec:
                responses_spec["200"] = OpenAPIResponse(
                    description="Successful Response",
                    content={
                        "application/json": MediaType(  # type:ignore
                            schema=Schema(
                                example={"example": "This is an example response"},
                                type="object",
                            )  # type:ignore
                        )
                    },
                )

            if responses:
                if isinstance(responses, type) and issubclass(responses, BaseModel):
                    self.add_schema(responses)
                elif hasattr(responses, "__origin__") and responses.__origin__ is list:
                    item_model = responses.__args__[0]
                    if issubclass(item_model, BaseModel):
                        self.add_schema(item_model)
                elif isinstance(responses, dict):
                    for model in responses.values():
                        if isinstance(model, type) and issubclass(model, BaseModel):
                            self.add_schema(model)
                        elif hasattr(model, "__origin__") and model.__origin__ is list:
                            item_model = model.__args__[0]
                            if issubclass(item_model, BaseModel):
                                self.add_schema(item_model)

            operation = Operation(
                summary=summary,
                description=description,
                responses=responses_spec,
                tags=tags,
                parameters=parameters or [],  # type:ignore
                requestBody=request_body_spec,
                security=security,
                operationId=operation_id or str(uuid4()),
                deprecated=deprecated,
                externalDocs=external_docs,
            )

            # Add operation to the OpenAPI specification
            if path not in self.config.openapi_spec.paths:
                self.config.openapi_spec.paths[path] = PathItem()

            setattr(self.config.openapi_spec.paths[path], method.lower(), operation)

            @wraps(func)
            async def wrapper(*args, **kwargs):
                return await func(*args, **kwargs)

            return wrapper

        return decorator

    def add_schema(self, schema: Type[BaseModel]) -> None:
        """Add a schema to the OpenAPI components section

        Args:
            schema: A Pydantic model to be added to the components/schemas section
        """
        if not self.config.openapi_spec.components:
            self.config.openapi_spec.components = Components()

        if not self.config.openapi_spec.components.schemas:
            self.config.openapi_spec.components.schemas = {}

        self.config.openapi_spec.components.schemas[schema.__name__] = Schema(
            **schema.model_json_schema()
        )



import os
import warnings
from pathlib import Path
from typing import List, Optional, Union

from nexios.http import Request, Response
from nexios.routing import BaseRouter
from nexios.types import Receive, Scope, Send


class StaticFilesHandler:
    def __init__(
        self,
        directory: Optional[Union[str, Path]] = None,
        directories: Optional[List[Union[str, Path]]] = None,
        url_prefix: str = "/static/",
    ):
        warnings.warn(
            "StaticFilesHandler is deprecated and will be removed in a future version. "
            "Please use StaticFiles instead with app.register(). Example:\n"
            "static_files = StaticFiles(directory='path/to/static')\n"
            "app.register(static_files, prefix='/static')",
            DeprecationWarning,
            stacklevel=2,
        )
        if directory is not None and directories is not None:
            raise ValueError("Cannot specify both 'directory' and 'directories'")
        if directory is None and directories is None:
            raise ValueError("Must specify either 'directory' or 'directories'")

        if directory is not None:
            self.directories = [self._ensure_directory(directory)]
        else:
            self.directories = [self._ensure_directory(d) for d in directories or []]

        self.url_prefix = url_prefix.rstrip("/") + "/"

    def _ensure_directory(self, path: Union[str, Path]) -> Path:
        """Ensure directory exists and return resolved Path"""
        directory = Path(path).resolve()
        if not directory.exists():
            os.makedirs(directory, exist_ok=True)
        if not directory.is_dir():
            raise ValueError(f"{directory} is not a directory")
        return directory

    def _is_safe_path(self, path: Path) -> bool:
        """Check if the path is safe to serve"""
        try:
            full_path = path.resolve()
            return any(
                str(full_path).startswith(str(directory))
                for directory in self.directories
            )
        except (ValueError, RuntimeError):
            return False

    async def __call__(self, request: Request, response: Response):
        path = request.path_params.get("path", "")

        if not path:
            return response.json("Invalid static file path", status_code=400)

        for directory in self.directories:
            try:
                file_path = (directory / path).resolve()

                if self._is_safe_path(file_path) and file_path.is_file():
                    return response.file(
                        str(file_path), content_disposition_type="inline"
                    )
            except (ValueError, RuntimeError):
                continue

        return response.json("Resource not found", status_code=404)


class StaticFiles(BaseRouter):
    def __init__(
        self,
        directory: Optional[Union[str, Path]] = None,
        directories: Optional[List[Union[str, Path]]] = None,
    ):
        if not directories:
            directories = [directory] if directory else []
        self.directories = [self._ensure_directory(d) for d in directories or []]

    def _ensure_directory(self, path: Union[str, Path]) -> Path:
        """Ensure directory exists and return resolved Path"""
        directory = Path(path).resolve()
        if not directory.exists():
            os.makedirs(directory, exist_ok=True)
        if not directory.is_dir():
            raise ValueError(f"{directory} is not a directory")
        return directory

    def _is_safe_path(self, path: Path) -> bool:
        """Check if the path is safe to serve"""
        try:
            full_path = path.resolve()
            return any(
                str(full_path).startswith(str(directory))
                for directory in self.directories
            )
        except (ValueError, RuntimeError):
            return False

    async def _handle(self, request: Request, response: Response):
        path = request.scope.get("path", "").lstrip("/")
        for directory in self.directories:
            try:
                file_path = (directory / path).resolve()
                if self._is_safe_path(file_path) and file_path.is_file():
                    return response.file(
                        str(file_path), content_disposition_type="inline"
                    )
            except (ValueError, RuntimeError):
                continue

        return response.json("Resource not found", status_code=404)

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        request = Request(scope, receive)
        response = Response(request)
        await self._handle(request, response)
        result = response.get_response()
        if result is not None:
            response = result
        await response(scope, receive, send)



import pytest
from nexios.views import APIView
from nexios.http import Request, Response
import types
import asyncio

class DummyRequest:
    def __init__(self, method="GET"):
        self.method = method
        self.url = types.SimpleNamespace(path="/test")

class DummyResponse:
    def __init__(self):
        self._status_code = 200
        self._json = None
    def status(self, code):
        self._status_code = code
        return self
    def json(self, data):
        self._json = data
        return self

@pytest.mark.asyncio
async def test_api_view_get_default():
    view = APIView()
    req = DummyRequest("GET")
    res = DummyResponse()
    response = await view.get(req, res)
    assert response._status_code == 404
    assert response._json == {"error": "Not Found"}

@pytest.mark.asyncio
async def test_api_view_method_not_allowed():
    view = APIView()
    req = DummyRequest("OPTIONS")
    res = DummyResponse()
    response = await view.method_not_allowed(req, res)
    assert response._status_code == 405
    assert response._json == {"error": "Method Not Allowed"}

@pytest.mark.asyncio
async def test_api_view_dispatch_calls_correct_method():
    class MyView(APIView):
        async def get(self, req, res):
            return res.status(200).json({"ok": True})
    view = MyView()
    req = DummyRequest("GET")
    res = DummyResponse()
    response = await view.dispatch(req, res)
    assert response._status_code == 200
    assert response._json == {"ok": True}

@pytest.mark.asyncio
async def test_api_view_dispatch_error_handler():
    class MyView(APIView):
        async def get(self, req, res):
            raise ValueError("fail")
    called = {}
    async def handler(req, res, exc):
        called["handled"] = True
        return res.status(400).json({"error": str(exc)})
    MyView.error_handlers = {ValueError: handler}
    view = MyView()
    req = DummyRequest("GET")
    res = DummyResponse()
    response = await view.dispatch(req, res)
    assert response._status_code == 400
    assert response._json == {"error": "fail"}
    assert called["handled"] 


from datetime import datetime

import humanize
import markdown2

from nexios import NexiosApp
from nexios.templating import TemplateConfig, render

app = NexiosApp()


def format_datetime(value, format="%Y-%m-%d %H:%M"):
    """Custom filter to format datetime objects."""
    if isinstance(value, str):
        value = datetime.fromisoformat(value)
    return value.strftime(format)


def humanize_number(value):
    """Custom filter to humanize numbers."""
    return humanize.intcomma(value)


def markdown_to_html(value):
    """Custom filter to convert markdown to HTML."""
    return markdown2.markdown(value)


# Configure templating with custom filters and globals
template_config = TemplateConfig(
    template_dir="templates",
    custom_filters={
        "datetime": format_datetime,
        "humanize": humanize_number,
        "markdown": markdown_to_html,
    },
    custom_globals={
        "site_name": "Nexios Demo",
        "current_year": datetime.now().year,
        "version": "1.0.0",
    },
)

app.config.templating = template_config


@app.get("/blog")
async def blog_list(request, response):
    """Example using custom filters and globals."""
    posts = [
        {
            "title": "Getting Started with Nexios",
            "content": """
# Getting Started
Nexios is a modern web framework that makes it easy to build web applications.
- Fast
- Secure
- Easy to use
            """,
            "views": 12345,
            "published": "2024-03-15T10:30:00",
        },
        {
            "title": "Advanced Templating",
            "content": """
# Advanced Templates
Learn how to use advanced templating features:
1. Custom filters
2. Global variables
3. Markdown support
            """,
            "views": 5432,
            "published": "2024-03-16T15:45:00",
        },
    ]

    return await render("blog/list.html", {"posts": posts})


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8000)

"""
Example template (blog/list.html):
{% extends 'base.html' %}

{% block content %}
    <h1>{{ site_name }} Blog</h1>
    
    {% for post in posts %}
    <article class="post">
        <h2>{{ post.title }}</h2>
        <div class="metadata">
            Published: {{ post.published|datetime('%B %d, %Y') }}
            <br>
            Views: {{ post.views|humanize }}
        </div>
        <div class="content">
            {{ post.content|markdown }}
        </div>
    </article>
    {% endfor %}
    
    <footer>
        <p>{{ site_name }} v{{ version }} - © {{ current_year }}</p>
    </footer>
{% endblock %}
"""



import glob
import json
import os
from datetime import datetime

import matplotlib.pyplot as plt
import pandas as pd


def calculate_avg_rps(num_reqs_per_sec):
    """Calculate average requests per second from the num_reqs_per_sec dictionary"""
    total_requests = sum(num_reqs_per_sec.values())
    if not num_reqs_per_sec:
        return 0
    return total_requests / len(num_reqs_per_sec)


def plot_results(df, timestamp):
    """Create plots from the benchmark results"""
    plots_dir = "plots"
    os.makedirs(plots_dir, exist_ok=True)

    # Set style
    plt.style.use("seaborn-v0_8-darkgrid")

    # Create figure with two subplots
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 12))

    # Plot 1: Requests per Second
    bars1 = ax1.bar(df["framework"], df["requests_per_second"], color="#2ecc71")
    ax1.set_title("Requests per Second by Framework", pad=20, fontsize=14)
    ax1.set_ylabel("Requests/Second", fontsize=12)
    ax1.grid(True, linestyle="--", alpha=0.7)

    # Add value labels on top of bars
    for bar in bars1:
        height = bar.get_height()
        ax1.text(
            bar.get_x() + bar.get_width() / 2.0,
            height,
            f"{height:.1f}",
            ha="center",
            va="bottom",
        )

    # Plot 2: Average Response Time
    bars2 = ax2.bar(df["framework"], df["avg_response_time"], color="#e74c3c")
    ax2.set_title("Average Response Time by Framework", pad=20, fontsize=14)
    ax2.set_ylabel("Response Time (ms)", fontsize=12)
    ax2.grid(True, linestyle="--", alpha=0.7)

    # Add value labels on top of bars
    for bar in bars2:
        height = bar.get_height()
        ax2.text(
            bar.get_x() + bar.get_width() / 2.0,
            height,
            f"{height:.1f}",
            ha="center",
            va="bottom",
        )

    # Adjust layout and save
    plt.tight_layout()
    plt.savefig(
        f"{plots_dir}/benchmark_results_{timestamp}.png", dpi=300, bbox_inches="tight"
    )
    plt.close()


def analyze_results():
    results_dir = "results"
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    summary_file = f"benchmark_summary_{timestamp}.csv"

    # Initialize data for summary
    summary_data = []

    # Process each framework's results
    for result_file in glob.glob(f"{results_dir}/*.json"):
        framework = os.path.basename(result_file).split("_")[0]

        with open(result_file, "r") as f:
            data = json.load(f)

            # Calculate total metrics across all endpoints
            total_requests = sum(endpoint["num_requests"] for endpoint in data)
            total_response_time = sum(
                endpoint["total_response_time"] for endpoint in data
            )

            # Calculate average RPS across all endpoints
            all_reqs_per_sec = {}
            for endpoint in data:
                for timestamp, count in endpoint["num_reqs_per_sec"].items():
                    all_reqs_per_sec[timestamp] = (
                        all_reqs_per_sec.get(timestamp, 0) + count
                    )

            avg_rps = calculate_avg_rps(all_reqs_per_sec)
            avg_response_time = (
                total_response_time / total_requests if total_requests > 0 else 0
            )

            summary_data.append(
                {
                    "framework": framework,
                    "total_requests": total_requests,
                    "requests_per_second": avg_rps,
                    "avg_response_time": avg_response_time,
                }
            )

    # Create DataFrame and save to CSV
    df = pd.DataFrame(summary_data)
    df.to_csv(summary_file, index=False)
    print(f"\nResults saved to {summary_file}")

    # Generate plots
    plot_results(df, timestamp)
    print(f"Plots saved to plots/benchmark_results_{timestamp}.png")


if __name__ == "__main__":
    analyze_results()



import time

from flask import Flask, jsonify, request

app = Flask(__name__)


@app.route("/")
def root():
    return jsonify({"message": "Hello World"})


@app.route("/json")
def json_endpoint():
    return jsonify(
        {
            "string": "Hello World",
            "number": 42,
            "boolean": True,
            "array": [1, 2, 3],
            "object": {"key": "value"},
        }
    )


@app.route("/echo", methods=["POST"])
def echo():
    return jsonify(request.get_json())


@app.route("/delay/<float:seconds>")
def delay(seconds):
    time.sleep(seconds)
    return jsonify({"message": f"Delayed for {seconds} seconds"})


if __name__ == "__main__":
    app.run(host="127.0.0.1", port=8001)



from .backends.apikey import APIKeyAuthBackend
from .backends.jwt import JWTAuthBackend, create_jwt, decode_jwt
from .decorator import auth
from .middleware import AuthenticationMiddleware

__all__ = [
    "AuthenticationMiddleware",
    "APIKeyAuthBackend",
    "JWTAuthBackend",
    "create_jwt",
    "decode_jwt",
    "auth",
]



import os
import subprocess
import sys
from pathlib import Path
from unittest.mock import MagicMock, Mock, patch

import click
import pytest

from nexios.cli import cli
from nexios.cli.commands import (
    new,
    run,
)
from nexios.cli.utils import (
    _check_server_installed,
    _echo_error,
    _echo_info,
    _echo_success,
    _echo_warning,
    _find_app_module,
    _has_write_permission,
    _is_port_in_use,
    _validate_app_path,
    _validate_host,
    _validate_port,
    _validate_project_name,
    _validate_project_title,
    _validate_server,
)


class TestCLIUtilityFunctions:
    """Test CLI utility functions"""

    def test_echo_success(self, capsys):
        _echo_success("Test success")
        captured = capsys.readouterr()
        assert "✓ Test success" in captured.out

    def test_echo_error(self, capsys):
        _echo_error("Test error")
        captured = capsys.readouterr()
        assert "✗ Test error" in captured.err

    def test_echo_info(self, capsys):
        _echo_info("Test info")
        captured = capsys.readouterr()
        assert "ℹ Test info" in captured.out

    def test_echo_warning(self, capsys):
        _echo_warning("Test warning")
        captured = capsys.readouterr()
        assert "⚠ Test warning" in captured.out

    def test_has_write_permission(self, tmp_path):
        test_file = tmp_path / "test.txt"
        test_file.write_text("test")
        assert _has_write_permission(test_file) is True

    @patch("socket.socket")
    def test_is_port_in_use_true(self, mock_socket):
        mock_socket.return_value.__enter__.return_value.connect_ex.return_value = 0
        assert _is_port_in_use("localhost", 8000) is True

    @patch("socket.socket")
    def test_is_port_in_use_false(self, mock_socket):
        mock_socket.return_value.__enter__.return_value.connect_ex.return_value = 1
        assert _is_port_in_use("localhost", 8000) is False

    @patch("subprocess.run")
    def test_check_server_installed_true(self, mock_run):
        mock_run.return_value.returncode = 0
        assert _check_server_installed("uvicorn") is True

    @patch("subprocess.run")
    def test_check_server_installed_false(self, mock_run):
        mock_run.side_effect = FileNotFoundError()
        assert _check_server_installed("nonexistent") is False

    @patch("subprocess.run")
    def test_check_server_installed_called_process_error(self, mock_run):
        mock_run.side_effect = subprocess.CalledProcessError(1, "test")
        assert _check_server_installed("test") is False


class TestCLIValidationFunctions:
    """Test CLI validation functions"""

    def test_validate_project_name_valid(self):
        ctx = Mock()
        param = Mock()
        result = _validate_project_name(ctx, param, "valid_project")
        assert result == "valid_project"

    def test_validate_project_name_invalid(self):
        ctx = Mock()
        param = Mock()
        with pytest.raises(click.BadParameter):
            _validate_project_name(ctx, param, "invalid-project")

    def test_validate_project_name_empty(self):
        ctx = Mock()
        param = Mock()
        result = _validate_project_name(ctx, param, "")
        assert result == ""

    def test_validate_project_title_valid(self):
        ctx = Mock()
        param = Mock()
        result = _validate_project_title(ctx, param, "Valid Title")
        assert result == "Valid Title"

    def test_validate_project_title_invalid(self):
        ctx = Mock()
        param = Mock()
        with pytest.raises(click.BadParameter):
            _validate_project_title(ctx, param, "Invalid@Title")

    def test_validate_host_valid(self):
        ctx = Mock()
        param = Mock()
        result = _validate_host(ctx, param, "localhost")
        assert result == "localhost"

    def test_validate_host_invalid(self):
        ctx = Mock()
        param = Mock()
        with pytest.raises(click.BadParameter):
            _validate_host(ctx, param, "invalid@host")

    def test_validate_port_valid(self):
        ctx = Mock()
        param = Mock()
        result = _validate_port(ctx, param, 8000)
        assert result == 8000

    def test_validate_port_invalid_low(self):
        ctx = Mock()
        param = Mock()
        with pytest.raises(click.BadParameter):
            _validate_port(ctx, param, 0)

    def test_validate_port_invalid_high(self):
        ctx = Mock()
        param = Mock()
        with pytest.raises(click.BadParameter):
            _validate_port(ctx, param, 70000)

    def test_validate_app_path_valid(self):
        ctx = Mock()
        param = Mock()
        result = _validate_app_path(ctx, param, "main:app")
        assert result == "main:app"

    def test_validate_app_path_invalid(self):
        ctx = Mock()
        param = Mock()
        with pytest.raises(click.BadParameter):
            _validate_app_path(ctx, param, "invalid:path:format")

    def test_validate_server_valid(self):
        ctx = Mock()
        param = Mock()
        result = _validate_server(ctx, param, "uvicorn")
        assert result == "uvicorn"

    def test_validate_server_invalid(self):
        ctx = Mock()
        param = Mock()
        with pytest.raises(click.BadParameter):
            _validate_server(ctx, param, "invalid_server")


class TestCLIHelperFunctions:
    """Test CLI helper functions"""

    def test_find_app_module_main_py(self, tmp_path):
        main_py = tmp_path / "main.py"
        main_py.write_text("app = None")
        result = _find_app_module(tmp_path)
        assert result == "main:app"

    def test_find_app_module_app_main_py(self, tmp_path):
        app_dir = tmp_path / "app"
        app_dir.mkdir()
        main_py = app_dir / "main.py"
        main_py.write_text("app = None")
        result = _find_app_module(tmp_path)
        assert result == "app.main:app"

    def test_find_app_module_src_main_py(self, tmp_path):
        src_dir = tmp_path / "src"
        src_dir.mkdir()
        main_py = src_dir / "main.py"
        main_py.write_text("app = None")
        result = _find_app_module(tmp_path)
        assert result == "src.main:app"

    def test_find_app_module_not_found(self, tmp_path):
        result = _find_app_module(tmp_path)
        assert result is None


class TestCLIGroup:
    """Test CLI group functionality"""

    def test_cli_group_creation(self):
        """Test that the CLI group can be created"""
        assert cli.name == "cli"
        assert len(cli.commands) > 0

    def test_cli_has_new_command(self):
        """Test that the CLI has the new command"""
        assert "new" in cli.commands

    def test_cli_has_run_command(self):
        """Test that the CLI has the run command"""
        assert "run" in cli.commands



from nexios import NexiosApp
from nexios.http import Request, Response

app = NexiosApp()


@app.get("/items/{item_id}")
async def get_item(req: Request, res: Response):
    item_id = req.path_params["item_id"]
    return {"item_id": item_id}


@app.post("/items/{item_id}")
async def create_item(req: Request, res: Response):
    item_id = req.path_params["item_id"]
    item_data = await req.json
    return res.json({"item_id": item_id, "data": item_data}, status_code=201)


@app.get("/products/{product_id:int}")
async def get_product(req: Request, res: Response):
    product_id = req.path_params["product_id"]
    return res.json({"product_id": product_id, "type": "integer"})


@app.get("/categories/{category_name:str}")
async def get_category(req: Request, res: Response):
    category_name = req.path_params["category_name"]
    return res.json({"category_name": category_name, "type": "string"})


@app.get("/prices/{price:float}")
async def get_price(req: Request, res: Response):
    price = req.path_params["price"]
    return res.json({"price": price, "type": "float"})


@app.get("/uuids/{uuid_value:uuid}")
async def get_uuid(req: Request, res: Response):
    uuid_value = req.path_params["uuid_value"]
    return res.json({"uuid": str(uuid_value), "type": "UUID"})


@app.get("/slugs/{slug_value:slug}")
async def get_slug(req: Request, res: Response):
    slug_value = req.path_params["slug_value"]
    return res.json({"slug": slug_value, "type": "slug"})


@app.get("/wildcard/{wildcard_path:path}")
async def get_wildcard(req: Request, res: Response):
    wildcard_path = req.path_params["wildcard_path"]
    return res.json({"wildcard_path": wildcard_path, "type": "path"})



import pytest

from nexios import get_application
from nexios.http import Request, Response
from nexios.pagination import AsyncListDataHandler as ListDataHandler
from nexios.pagination import (
    AsyncPaginator,
    CursorPagination,
    InvalidCursorError,
    InvalidPageError,
    InvalidPageSizeError,
    LimitOffsetPagination,
    PageNumberPagination,
    PaginationError,
)
from nexios.testing import Client


@pytest.fixture
async def test_client():
    app = get_application()
    async with Client(app) as client:
        yield client, app


# Sample test data
TEST_DATA = [{"id": i, "name": f"Item {i}"} for i in range(1, 101)]


async def test_page_number_pagination_in_app(test_client):
    client, app = test_client

    @app.get("/items")
    async def get_items(req: Request, res: Response):
        handler = ListDataHandler(TEST_DATA)
        pagination = PageNumberPagination()
        base_url = str(req.url)

        try:
            paginator = AsyncPaginator(
                handler, pagination, base_url, dict(req.query_params)
            )
            result = await paginator.paginate()
            return res.json(result)
        except PaginationError as e:
            return res.json({"error": str(e)}, status_code=400)

    # Test basic pagination
    response = await client.get("/items?page=2&page_size=10")
    data = response.json()
    assert response.status_code == 200
    assert len(data["items"]) == 10
    assert data["items"][0]["id"] == 11
    assert data["pagination"]["page"] == 2
    assert data["pagination"]["total_items"] == 100
    assert "next" in data["pagination"]["links"]
    assert "prev" in data["pagination"]["links"]

    # Test invalid page
    response = await client.get("/items?page=0")
    assert response.status_code == 400
    assert "Page number must be at least 1" in response.json()["error"]

    # Test max page size
    response = await client.get("/items?page_size=200")
    data = response.json()
    assert data["pagination"]["page_size"] == 100  # Max page size enforced


async def test_limit_offset_pagination_in_app(test_client):
    client, app = test_client

    @app.get("/items-limit-offset")
    async def get_items(req: Request, res: Response):
        handler = ListDataHandler(TEST_DATA)
        pagination = LimitOffsetPagination()
        base_url = str(req.url)

        try:
            paginator = AsyncPaginator(
                handler, pagination, base_url, dict(req.query_params)
            )
            result = await paginator.paginate()
            return res.json(result)
        except PaginationError as e:
            return res.json({"error": str(e)}, status_code=400)

    # Test basic pagination
    response = await client.get("/items-limit-offset?limit=15&offset=30")
    data = response.json()
    assert response.status_code == 200
    assert len(data["items"]) == 15
    assert data["items"][0]["id"] == 31
    assert data["pagination"]["offset"] == 30
    assert data["pagination"]["total_items"] == 100
    assert "next" in data["pagination"]["links"]
    assert "prev" in data["pagination"]["links"]

    # Test invalid offset
    response = await client.get("/items-limit-offset?offset=-1")
    assert response.status_code == 400
    assert "Offset cannot be negative" in response.json()["error"]


async def test_cursor_pagination_in_app(test_client):
    client, app = test_client

    @app.get("/items-cursor")
    async def get_items(req: Request, res: Response):
        handler = ListDataHandler(TEST_DATA)
        pagination = CursorPagination()
        base_url = str(req.url)

        try:
            paginator = AsyncPaginator(
                handler, pagination, base_url, dict(req.query_params)
            )
            result = await paginator.paginate()
            return res.json(result)
        except PaginationError as e:
            return res.json({"error": str(e)}, status_code=400)

    # Test initial request
    response = await client.get("/items-cursor?page_size=10")
    data = response.json()
    assert response.status_code == 200
    assert len(data["items"]) == 10
    assert "next" in data["pagination"]["links"]
    assert "prev" not in data["pagination"]["links"]  # No prev on first page

    # Test with cursor
    next_cursor = data["pagination"]["links"]["next"].split("cursor=")[1].split("&")[0]
    response = await client.get(f"/items-cursor?cursor={next_cursor}&page_size=10")
    data = response.json()
    assert response.status_code == 200
    assert len(data["items"]) == 10
    # assert data["items"][0]["id"] == 11
    assert "next" in data["pagination"]["links"]
    assert "prev" in data["pagination"]["links"]

    # Test invalid cursor
    response = await client.get("/items-cursor?cursor=invalid")
    # assert response.status_code == 400
    # assert "Invalid cursor format" in response.json()["error"]


async def test_pagination_with_filters(test_client):
    client, app = test_client

    # Create filtered data handler
    class FilteredDataHandler(ListDataHandler):
        async def get_total_items(self) -> int:
            return len([item for item in self.data if item["id"] % 2 == 0])

        async def get_items(self, offset: int, limit: int) -> list:
            filtered = [item for item in self.data if item["id"] % 2 == 0]
            return filtered[offset : offset + limit]

    @app.get("/filtered-items")
    async def get_filtered_items(req: Request, res: Response):
        handler = FilteredDataHandler(TEST_DATA)
        pagination = PageNumberPagination()
        base_url = str(req.url)

        try:
            paginator = AsyncPaginator(
                handler, pagination, base_url, dict(req.query_params)
            )
            result = await paginator.paginate()
            return res.json(result)
        except PaginationError as e:
            return res.json({"error": str(e)}, status_code=400)

    response = await client.get("/filtered-items?page=2&page_size=10")
    data = response.json()
    assert response.status_code == 200
    assert len(data["items"]) == 10
    assert all(item["id"] % 2 == 0 for item in data["items"])
    assert data["pagination"]["total_items"] == 50  # Only even IDs


async def test_pagination_error_handling(test_client):
    client, app = test_client

    @app.get("/error-test")
    async def error_test(req: Request, res: Response):
        handler = ListDataHandler([])
        pagination = PageNumberPagination()
        base_url = str(req.url)

        try:
            paginator = AsyncPaginator(
                handler,
                pagination,
                base_url,
                dict(req.query_params),
                validate_total_items=False,
            )
            result = await paginator.paginate()
            return res.json(result)
        except InvalidPageError as e:
            return res.json({"error": str(e)}, status_code=400)
        except InvalidPageSizeError as e:
            return res.json({"error": str(e)}, status_code=400)
        except InvalidCursorError as e:
            return res.json({"error": str(e)}, status_code=400)
        except PaginationError as e:
            return res.json({"error": str(e)}, status_code=500)

    # Test various error cases
    response = await client.get("/error-test?page=0")
    assert response.status_code == 400
    assert "Page number must be at least 1" in response.json()["error"]

    response = await client.get("/error-test?page_size=0")
    assert response.status_code == 400
    assert "Page size must be at least 1" in response.json()["error"]


async def test_pagination_with_custom_metadata(test_client):
    client, app = test_client

    class CustomPagination(PageNumberPagination):
        def generate_metadata(self, total_items, items, base_url, request_params):
            metadata = super().generate_metadata(
                total_items, items, base_url, request_params
            )
            metadata["custom_field"] = "custom_value"
            return metadata

    @app.get("/custom-metadata")
    async def custom_metadata(req: Request, res: Response):
        handler = ListDataHandler(TEST_DATA)
        pagination = CustomPagination()
        base_url = str(req.url)

        paginator = AsyncPaginator(
            handler, pagination, base_url, dict(req.query_params)
        )
        result = await paginator.paginate()
        return res.json(result)

    response = await client.get("/custom-metadata?page=1")
    data = response.json()
    assert response.status_code == 200
    assert data["pagination"]["custom_field"] == "custom_value"



# test_config.py

import pytest

from nexios.config import MakeConfig, get_config, set_config


# Test MakeConfig class
def test_makeconfig_initialization():
    # Test basic initialization
    config = MakeConfig({"debug": True, "db": {"host": "localhost"}})
    assert config.debug is True
    assert config.db.host == "localhost"

    # Test with defaults
    config = MakeConfig(
        {"debug": False}, defaults={"debug": True, "db": {"port": 5432}}
    )
    assert config.debug is False  # Overrides default
    assert config.db.port == 5432  # From defaults

    # Test immutable config
    config = MakeConfig({"debug": True}, immutable=True)
    with pytest.raises(AttributeError):
        config.debug = False


def test_makeconfig_nested_access():
    config = MakeConfig({"app": {"name": "MyApp", "settings": {"timeout": 30}}})

    # Test attribute access
    assert config.app.name == "MyApp"
    assert config.app.settings.timeout == 30

    # Test dictionary-style access
    assert config["app.name"] == "MyApp"
    assert config["app.settings.timeout"] == 30
    assert config["nonexistent.key"] is None


def test_makeconfig_conversion_methods():
    config = MakeConfig({"debug": True, "db": {"host": "localhost", "port": 5432}})

    # Test to_dict
    config_dict = config.to_dict()
    assert config_dict == {"debug": True, "db": {"host": "localhost", "port": 5432}}

    # Test to_json
    config_json = config.to_json()
    assert '"debug": true' in config_json
    assert '"host": "localhost"' in config_json


def test_makeconfig_repr():
    config = MakeConfig({"debug": True})
    assert repr(config) == "MakeConfig({'debug': True})"


def test_config_immutability():
    # Create mutable config
    config = MakeConfig({"debug": True})
    set_config(config)

    # Should be able to modify
    get_config().debug = False
    assert get_config().debug is False

    # Create immutable config
    immutable_config = MakeConfig({"debug": True}, immutable=True)
    set_config(immutable_config)

    # Should not be able to modify
    with pytest.raises(AttributeError):
        get_config().debug = False


# Test edge cases
def test_makeconfig_edge_cases():
    # Test empty config
    empty_config = MakeConfig({})
    assert empty_config.to_dict() == {}

    # Test None values
    config_with_none = MakeConfig({"debug": None})
    assert config_with_none.debug is None

    # Test non-dict values
    config = MakeConfig({"version": "1.0"})
    assert config.version == "1.0"


def test_makeconfig_nested_immutability():
    # Test nested immutability
    config = MakeConfig({"db": {"host": "localhost"}}, immutable=True)

    with pytest.raises(AttributeError):
        config.db.host = "127.0.0.1"



from nexios import NexiosApp
from nexios.websockets import WebSocket

app = NexiosApp()


@app.ws_route("/ws")
async def websocket_handler(websocket: WebSocket):
    await websocket.accept()

    try:
        while True:
            message = await websocket.receive_text()
            await websocket.send_text(f"Echo: {message}")
    except Exception as e:
        print(f"WebSocket error: {e}")
    finally:
        await websocket.close()



from nexios import NexiosApp
from nexios.http.response import (
    BaseResponse,
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
)

app = NexiosApp()


@app.get("/json")
async def json_handler(req, res):
    return JSONResponse({"message": "Hello, World!"})


@app.get("/html")
async def html_handler(req, res):
    return HTMLResponse("<h1>Hello, World!</h1>")


@app.get("/text")
async def text_handler(req, res):
    return PlainTextResponse("Hello, World!", content_type="text/plain")


@app.get("/file")
async def file_handler(req, res):
    return FileResponse(
        "examples/respons/example_004.py", content_disposition_type="attachment"
    )


@app.get("/raw")
async def raw_handler(req, res):
    return BaseResponse(b"Hello, World!", content_type="text/plain")



from __future__ import annotations

import sys
from logging import (
    DEBUG,
    Formatter,
    Handler,
    Logger,
    LogRecord,
    StreamHandler,
    getLogger,
)
from logging.handlers import QueueHandler, QueueListener, RotatingFileHandler
from queue import SimpleQueue as Queue
from typing import TYPE_CHECKING, Optional, Tuple

if TYPE_CHECKING:
    pass


class LocalQueueHandler(QueueHandler):
    """Custom QueueHandler to avoid unnecessary record preparation.

    Since we use an in-process queue, there's no need to prepare records,
    reducing logging overhead.
    """

    def prepare(self, record: LogRecord) -> LogRecord:
        return record


def _setup_logging_queue(*handlers: Handler) -> QueueHandler:
    """Creates a LocalQueueHandler and starts a QueueListener."""
    queue: Queue[LogRecord] = Queue()
    queue_handler = LocalQueueHandler(queue)

    listener = QueueListener(queue, *handlers, respect_handler_level=True)
    listener.start()

    return queue_handler


def has_level_handler(logger: Logger) -> bool:
    """Checks if the logger already has an appropriate handler."""
    level = logger.getEffectiveLevel()
    current_logger: Optional[Logger] = logger

    while current_logger:
        if any(handler.level <= level for handler in current_logger.handlers):
            return True
        if not current_logger.propagate:
            break
        current_logger = current_logger.parent

    return False


def create_logger(
    logger_name: str = "nexios",
    log_level: int = DEBUG,
    log_file: Optional[str] = None,
    max_bytes: int = 10 * 1024 * 1024,  # 10MB per log file
    backup_count: int = 5,
) -> Logger:
    """Creates a high-performance, configurable logger for Nexios.

    Args:
        logger_name (str): The name of the logger.
        log_level (int): The logging level (DEBUG, INFO, ERROR, etc.).
        log_file (Optional[str]): Path to a log file for file-based logging.
        max_bytes (int): Max size of each log file before rotating.
        backup_count (int): Number of backup log files to keep.

    Returns:
        Logger: A configured logger instance.
    """
    logger = getLogger(logger_name)
    logger.setLevel(log_level)

    console_handler = StreamHandler(sys.stderr)
    console_handler.setFormatter(
        Formatter("[%(asctime)s] %(levelname)s in %(module)s: %(message)s")
    )

    handlers: Tuple[Handler, ...] = (console_handler,)

    if log_file:
        file_handler = RotatingFileHandler(
            log_file, maxBytes=max_bytes, backupCount=backup_count
        )
        file_handler.setFormatter(
            Formatter("[%(asctime)s] %(levelname)s in %(module)s: %(message)s")
        )
        handlers += (file_handler,)

    if not has_level_handler(logger):
        queue_handler = _setup_logging_queue(*handlers)
        logger.addHandler(queue_handler)

    return logger



#!/usr/bin/env python
"""
Nexios CLI - New project command.
"""

import sys
from pathlib import Path
from typing import Optional

import click

from nexios.__main__ import __version__

from ..utils import (
    _echo_error,
    _echo_info,
    _echo_success,
    _echo_warning,
    _has_write_permission,
    _validate_project_name,
    _validate_project_title,
)


@click.command()
@click.argument("project_name", callback=_validate_project_name, required=True)
@click.option(
    "--output-dir",
    "-o",
    default=".",
    help="Directory where the project should be created.",
    type=click.Path(file_okay=False),
)
@click.option(
    "--title",
    help="Display title for the project (defaults to project name if not provided).",
    callback=_validate_project_title,
)
@click.option(
    "--template",
    "-t",
    type=click.Choice(["basic", "standard", "beta"], case_sensitive=False),
    default="basic",
    help="Template type to use for the project.",
    show_default=True,
)
def new(
    project_name: str,
    output_dir: str,
    title: Optional[str] = None,
    template: str = "basic",
):
    """
    Create a new Nexios project.

    Creates a new Nexios project with the given name in the specified directory.
    The project will be initialized with the selected template structure including
    configuration files and a main application file.

    Available template types:
    - basic: Minimal starter template with essential structure
    - standard: A complete template with commonly used features
    - beta: An advanced template with experimental features
    """
    try:
        output_path = Path(output_dir).resolve()
        project_path = output_path / project_name

        if not project_name.strip():
            _echo_error("Project name cannot be empty.")
            return

        if project_path.exists():
            _echo_error(
                f"Directory {project_path} already exists. Choose a different name or location."
            )
            return

        if not _has_write_permission(output_path):
            _echo_error(
                f"No write permission for directory {output_path}. Choose a different location or run with appropriate permissions."
            )
            return

        project_path.mkdir(parents=True, exist_ok=True)
        _echo_info(
            f"Creating new Nexios project: {project_name} using {template} template"
        )

        template_dir = (
            Path(__file__).parent.parent.parent / "templates" / template.lower()
        )

        if not template_dir.exists():
            _echo_error(
                f"Template directory for '{template}' not found: {template_dir}"
            )
            _echo_error(
                "Please ensure you have the latest version of Nexios installed."
            )
            available_templates = [
                p.name
                for p in (Path(__file__).parent.parent.parent / "templates").glob("*")
                if p.is_dir()
            ]
            if available_templates:
                _echo_info(f"Available templates: {', '.join(available_templates)}")
            return

        for src_path in template_dir.glob("**/*"):
            if src_path.is_dir():
                continue

            rel_path = src_path.relative_to(template_dir)
            dest_path = project_path / rel_path
            dest_path.parent.mkdir(parents=True, exist_ok=True)

            try:
                content = src_path.read_text(encoding="utf-8")
                project_title = title or project_name.replace("_", " ").title()
                content = content.replace("{{project_name}}", project_name)
                content = content.replace("{{project_name_title}}", project_title)
                content = content.replace("{{version}}", __version__)
                dest_path.write_text(content, encoding="utf-8")

            except PermissionError:
                _echo_error(
                    f"Permission denied when writing to {dest_path}. Please check your file permissions."
                )
                return
            except Exception as e:
                _echo_warning(f"Error processing template file {src_path}: {str(e)}")

        env_path = project_path / ".env"
        env_content = [
            "# Environment variables for the Nexios application",
            "DEBUG=True",
            "HOST=127.0.0.1",
            "PORT=4000",
        ]
        env_path.write_text("\n".join(env_content) + "\n", encoding="utf-8")

        TEMPLATE_CONFIG = """\
from nexios import NexiosApp

app = NexiosApp()

cli_options = {
    "host": "127.0.0.1",
    "port": 4000,
    "reload": True,
    "server": "uvicorn",
}

# Optionally, you can set a custom_command to start your app
# custom_command = "gunicorn -w 4 -b 0.0.0.0:4000 nexios.config:app"
"""

        config_path = project_path / "nexios.config.py"
        config_path.write_text(TEMPLATE_CONFIG, encoding="utf-8")

        _echo_success(f"Project {project_name} created successfully at {project_path}")
        _echo_info("Next steps:")
        _echo_info(f"  1. cd {project_name}")
        _echo_info("  2. pip install -r requirements.txt")
        _echo_info("  3. nexios run")

    except Exception as e:
        _echo_error(f"Error creating project: {str(e)}")
        sys.exit(1)



from typing import Any, Callable

from nexios.auth.base import AuthenticationBackend, UnauthenticatedUser
from nexios.http import Request, Response


class SessionAuthBackend(AuthenticationBackend):
    """
    Session-based authentication backend that integrates with the framework's
    built-in session manager (req.session).

    This backend checks for authenticated user data in the existing session.
    """

    def __init__(
        self,
        authenticate_func: Callable[..., Any],
        user_key: str = "user",
    ):
        """
        Initialize the session auth backend.

        Args:
            user_key (str): The key used to store user data in the session (default: "user")
        """
        self.user_key = user_key
        self.authenticate_func = authenticate_func

    async def authenticate(self, request: Request, response: Response) -> Any:
        """
        Authenticate the user using the framework's session.

        Args:
            request: The HTTP request containing the session
            response: The HTTP response (unused in this backend)

        Returns:
            Tuple[user_data, "session"] if authenticated
            None if no session exists
            UnauthenticatedUser if session exists but is invalid
        """
        assert "session" in request.scope, "No Session Middleware Installed"
        user_data = request.session.get(self.user_key)
        if not user_data:
            return None

        user = await self.authenticate_func(**user_data)
        if not user:
            return UnauthenticatedUser()
        return user, "session"



from datetime import datetime

from nexios.http import Request, Response
from nexios.middleware.base import BaseMiddleware


class ComplexMiddleware(BaseMiddleware):
    async def process_request(self, req: Request, res: Response, cnext):
        # Pre-processing: Log request details
        print(f"Request received: {req.method} {req.url}")

        # Authentication check
        if not req.headers.get("Authorization"):
            return res.json({"error": "Unauthorized"}, status_code=401)

        # Store request time
        req.state.request_time = datetime.now()

        # Proceed to next middleware or handler
        await cnext(req, res)

    async def process_response(self, req: Request, res: Response):
        # Post-processing: Add response headers
        res.headers["X-Processed-Time"] = str(datetime.now() - req.state.request_time)

        # Log response status
        print(f"Response sent with status: {res.status_code}")

        return res  # Return the modified response



from typing import Callable

from nexios.file_router.html import render
from nexios.http import Request, Response


@render("temps.html")
async def get(req: Request, res: Response):
    return {}


async def post_middleware(req: Request, res: Response, next: Callable):
    print(f"Received request: {req.method} {req.path}")
    await next()
    # raise Exception("AuthenticationRequired")


async def post(req: Request, res: Response):
    res.json({"ok": True})



from typing import Tuple

import pytest

from nexios import NexiosApp, get_application
from nexios.config import MakeConfig
from nexios.http import Request, Response
from nexios.session.file import FileSessionManager
from nexios.session.middleware import SessionMiddleware
from nexios.session.signed_cookies import SignedSessionManager
from nexios.testing import Client


# Fixtures for different session configurations
@pytest.fixture
async def file_session_client(tmp_path) -> Tuple[Client, NexiosApp]:
    """Client with file-based session configuration"""
    app = get_application(
        MakeConfig(
            {
                "secret_key": "file_session_secret",
                "session": {
                    "session_cookie_name": "file_session",
                    "session_permanent": True,
                    "session_expiration_time": 30,
                    "manager": FileSessionManager,
                },
                "session_file_name": str(tmp_path / "sessions"),
                "SESSION_FILE_STORAGE_PATH": str(tmp_path / "sessions"),
            }
        )
    )
    app.add_middleware(SessionMiddleware())
    async with Client(app) as client:
        yield client, app


@pytest.fixture
async def signed_session_client() -> Tuple[Client, NexiosApp]:
    """Client with signed cookie session configuration"""
    app = get_application(
        MakeConfig(
            {
                "secret_key": "signed_session_secret",
                "session": {
                    "session_cookie_name": "signed_session",
                    "session_permanent": True,
                    "session_expiration_time": 30,
                    "manager": SignedSessionManager,
                },
            }
        )
    )
    async with Client(app) as client:
        yield client, app


# # Test session middleware
async def test_session_middleware_initialization(
    file_session_client: Tuple[Client, NexiosApp],
):
    client, app = file_session_client

    @app.get("/test-session")
    async def test_session(req: Request, res: Response):

        req.session["key"] = "value"
        return res.text("OK")

    response = await client.get("/test-session")
    assert response.status_code == 200
    assert response.text == "OK"


async def test_session_middleware_no_secret_key():
    app = get_application(MakeConfig({"secret_key": None}))

    @app.get("/test-session")
    async def test_session(req: Request, res: Response):
        try:
            req.session["key"] = "value"
        except AssertionError:
            return res.text("No session", status_code=512)
        return res.text("OK")

    async with Client(app) as client:
        response = await client.get("/test-session")
        assert response.status_code == 512
        assert response.text == "No session"


# # Test file session manager
# async def test_file_session_operations(
#     file_session_client: Tuple[Client, NexiosApp], tmp_path
# ):
#     client, app = file_session_client

#     @app.get("/set-session")
#     async def set_session(req: Request, res: Response):
#         req.session["test_key"] = "test_value"
#         req.session["user"] = {"id": 1, "name": "Test"}
#         return res.text("Session set")

#     @app.get("/get-session")
#     async def get_session(req: Request, res: Response):
#         return res.json(
#             {"test_key": req.session.get("test_key"), "user": req.session.get("user")}
#         )

#     @app.get("/delete-session")
#     async def delete_session(req: Request, res: Response):
#         del req.session["test_key"]
#         return res.text("Session deleted")

#     @app.get("/clear-session")
#     async def clear_session(req: Request, res: Response):
#         req.session.clear()
#         return res.text("Session cleared")

#     #     # Set session
#     response = await client.get("/set-session")
#     assert response.status_code == 200

#     # Verify cookie was set
#     assert "file_session" in response.cookies

#     session_id = response.cookies["file_session"]

#     # # Verify session file was created
#     session_file = tmp_path / "sessions" / f"{session_id}.json"
#     assert session_file.exists()

#     response = await client.get("/get-session")
#     assert response.status_code == 200
#     # assert response.json() == {
#     #     "test_key": "test_value",
#     #     "user": {"id": 1, "name": "Test"},
#     # }

#     # Delete item from session
#     response = await client.get("/delete-session")
#     # assert response.status_code == 200

#     # Verify deletion
#     response = await client.get("/get-session")
#     # assert response.json()["test_key"] is None

#     # Clear session
#     response = await client.get("/clear-session")
#     # assert response.status_code == 200
#     # assert not session_file.exists()


# # Test signed cookie session manager
# async def test_signed_session_operations(signed_session_client: Tuple[Client, NexiosApp]):
#     client, app = signed_session_client

#     @app.get("/set-session")
#     async def set_session(req: Request, res: Response):
#         req.session["test_key"] = "test_value"
#         req.session["user"] = {"id": 1, "name": "Test"}
#         return res.text("Session set")

#     @app.get("/get-session")
#     async def get_session(req: Request, res: Response):
#         return res.json({
#             "test_key": req.session.get("test_key"),
#             "user": req.session.get("user")
#         })

#     # Set session
#     response = await client.get("/set-session")
#     assert response.status_code == 200

#     # Verify cookie was set
#     assert "signed_session" in response.cookies
#     session_cookie = response.cookies["signed_session"]

#     # Get session
#     response = await client.get("/get-session")
#     assert response.status_code == 200
#     assert response.json() == {
#         "test_key": "test_value",
#         "user": {"id": 1, "name": "Test"}
#     }

#     # Test with invalid cookie
#     client.cookies["signed_session"] = "invalid.token"
#     response = await client.get("/get-session")
#     assert response.status_code == 200
#     assert response.json() == {
#         "test_key": None,
#         "user": None
#     }

# # Test session expiration
# async def test_session_expiration(file_session_client: Tuple[Client, NexiosApp]):
#     client, app = file_session_client

#     @app.get("/set-expiring-session")
#     async def set_expiring_session(req: Request, res: Response):
#         req.session["temp"] = "data"
#         # Set expiration to 1 second from now
#         req.session._session_cache["__expires"] = (
#             datetime.now(timezone.utc) + timedelta(seconds=1)
#         ).isoformat()
#         return res.text("Session set")

#     @app.get("/check-expired")
#     async def check_expired(req: Request, res: Response):
#         return res.json({"expired": req.session.has_expired()})

#     # Set session
#     response = await client.get("/set-expiring-session")
#     assert response.status_code == 200

#     # Check immediately - should not be expired
#     response = await client.get("/check-expired")
#     assert response.status_code == 200
#     assert response.json()["expired"] is False

#     # Wait for expiration
#     import time
#     time.sleep(2)

#     # Check again - should be expired
#     response = await client.get("/check-expired")
#     assert response.status_code == 200
#     assert response.json()["expired"] is True

# # Test session cookie settings
# async def test_session_cookie_settings(file_session_client: Tuple[Client, NexiosApp]):
#     client, app = file_session_client

#     # Update cookie settings
#     app.config.session.session_cookie_httponly = True
#     app.config.session.session_cookie_secure = True
#     app.config.session.session_cookie_samesite = "lax"
#     app.config.session.session_cookie_path = "/test"
#     app.config.session.session_cookie_domain = "example.com"

#     @app.get("/set-cookie-settings")
#     async def set_cookie_settings(req: Request, res: Response):
#         req.session["test"] = "value"
#         return res.text("OK")

#     response = await client.get("/set-cookie-settings")
#     assert response.status_code == 200


#     cookie = response.cookies["file_session"]
#     assert cookie["httponly"] is True
#     assert cookie["secure"] is True
#     assert cookie["samesite"] == "lax"
#     assert cookie["path"] == "/test"
#     assert cookie["domain"] == "example.com"
# async def test_signed_session_operations(
#     signed_session_client: Tuple[Client, NexiosApp],
# ):
#     client, app = signed_session_client

#     @app.get("/set-session")
#     async def set_session(req: Request, res: Response):
#         req.session["test_key"] = "test_value"
#         req.session["user"] = {"id": 1, "name": "Test"}
#         return res.text("Session set")

#     @app.get("/get-session")
#     async def get_session(req: Request, res: Response):
#         return res.json(
#             {"test_key": req.session.get("test_key"), "user": req.session.get("user")}
#         )

#     # Set session
#     response = await client.get("/set-session")
#     assert response.status_code == 200

#     # Verify cookie was set
#     assert "signed_session" in response.cookies
#     session_cookie = response.cookies["signed_session"]

#     # Get session
#     response = await client.get("/get-session")
#     assert response.status_code == 200
#     assert response.json() == {
#         "test_key": "test_value",
#         "user": {"id": 1, "name": "Test"},
#     }

#     # Test with invalid cookie
#     client.cookies["signed_session"] = "invalid.token"
#     response = await client.get("/get-session")
#     assert response.status_code == 200
#     assert response.json() == {"test_key": None, "user": None}


# Test session cookie settings
async def test_session_cookie_settings(file_session_client: Tuple[Client, NexiosApp]):
    client, app = file_session_client

    # Update cookie settings
    app.config.session.session_cookie_httponly = True
    app.config.session.session_permanent = False

    @app.get("/set-cookie-settings")
    async def set_cookie_settings(req: Request, res: Response):
        req.session["test"] = "value"
        return res.text("OK")

    response = await client.get("/set-cookie-settings")
    assert response.status_code == 200

    response.cookies["file_session"]
    # assert cookie["httponly"] is True
    # assert cookie["secure"] is True
    # assert cookie["samesite"] == "lax"
    # assert cookie["path"] == "/test"
    # assert cookie["domain"] == "example.com"


# # Test session middleware with custom manager
# async def test_custom_session_manager(file_session_client: Tuple[Client, NexiosApp]):
#     # Define a simple in-memory session manager for testing
#     class MemorySessionManager(BaseSessionInterface):
#         _store: Dict[str, Dict[str, Any]] = {}

#         async def load(self):
#             self._session_cache = self._store.get(self.session_key, {})

#         async def save(self):
#             self._store[self.session_key] = self._session_cache

#     # app = get_application(MakeConfig({
#     #     "secret_key": "custom_session_secret",
#     #     "session": {
#     #         "manager": MemorySessionManager,
#     #         "session_cookie_name": "custom_session"
#     #     }
#     # }))
#     client, app = file_session_client
#     app.config.session.manager = MemorySessionManager
#     app.config.session.session_cookie_name = "custom"

#     @app.get("/test-custom-manager")
#     async def test_custom_manager(req: Request, res: Response):

#         if "count" not in req.session:
#             req.session["count"] = 1
#         else:

#             req.session["count"] += 1
#         return res.json({"count": req.session["count"]})

#     async with Client(app) as client:
#         # First request
#         response = await client.get("/test-custom-manager")
#         assert response.status_code == 200
#         assert response.json()["count"] == 1

#         # Second request
#         response = await client.get("/test-custom-manager")
#         assert response.status_code == 200
#         assert response.json()["count"] == 2

#         # New client should start fresh
#     async with Client(app) as new_client:
#         response = await new_client.get("/test-custom-manager")
#         assert response.status_code == 200
#         assert response.json()["count"] == 1



#!/usr/bin/env python
"""
Nexios CLI - Command line interface for the Nexios framework.

This module provides CLI commands for bootstrapping and running Nexios projects.
"""

from nexios.cli.main import cli

if __name__ == "__main__":
    cli()



from __future__ import annotations

import sys
import typing
from collections.abc import Iterator
from typing import Any, Protocol

import anyio

from nexios.dependencies import Context, current_context
from nexios.http.request import ClientDisconnect, Request
from nexios.http.response import NexiosResponse as Response
from nexios.types import ASGIApp, Message, Receive, Scope, Send
from nexios.utils.async_helpers import collapse_excgroups
from nexios.websockets import WebSocket

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

RequestResponseEndpoint = typing.Callable[[Request], typing.Awaitable[Response]]
DispatchFunction = typing.Callable[
    [Request, Response, typing.Callable[[], typing.Awaitable[Response]]],
    typing.Awaitable[Response],
]
T = typing.TypeVar("T")
P = ParamSpec("P")


class _MiddlewareFactory(Protocol[P]):
    def __call__(
        self, app: ASGIApp, /, *args: P.args, **kwargs: P.kwargs
    ) -> ASGIApp: ...  # pragma: no cover


class DefineMiddleware:
    def __init__(
        self,
        cls: _MiddlewareFactory[P],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.cls = cls
        self.args = args
        self.kwargs = kwargs

    def __iter__(self) -> Iterator[Any]:
        as_tuple = (self.cls, self.args, self.kwargs)
        return iter(as_tuple)

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        args_strings = [f"{value!r}" for value in self.args]
        option_strings = [f"{key}={value!r}" for key, value in self.kwargs.items()]
        name = getattr(self.cls, "__name__", "")
        args_repr = ", ".join([name] + args_strings + option_strings)
        return f"{class_name}({args_repr})"


class _CachedRequest(Request):
    """
    If the user calls Request.body() from their dispatch function
    we cache the entire request body in memory and pass that to downstream middleware,
    but if they call Request.stream() then all we do is send an
    empty body so that downstream things don't hang forever.
    """

    def __init__(self, scope: Scope, receive: Receive):
        super().__init__(scope, receive)
        self._wrapped_rcv_disconnected = False
        self._wrapped_rcv_consumed = False
        self._wrapped_rc_stream = self.stream()

    async def wrapped_receive(self) -> Message:
        # wrapped_rcv state 1: disconnected
        if self._wrapped_rcv_disconnected:
            # we've already sent a disconnect to the downstream app
            # we don't need to wait to get another one
            # (although most ASGI servers will just keep sending it)
            return {"type": "http.disconnect"}
        # wrapped_rcv state 1: consumed but not yet disconnected
        if self._wrapped_rcv_consumed:
            # since the downstream app has consumed us all that is left
            # is to send it a disconnect
            if self._is_disconnected:
                # the middleware has already seen the disconnect
                # since we know the client is disconnected no need to wait
                # for the message
                self._wrapped_rcv_disconnected = True
                return {"type": "http.disconnect"}
            # we don't know yet if the client is disconnected or not
            # so we'll wait until we get that message
            msg = await self.receive()
            if msg["type"] != "http.disconnect":  # pragma: no cover
                # at this point a disconnect is all that we should be receiving
                # if we get something else, things went wrong somewhere
                raise RuntimeError(f"Unexpected message received: {msg['type']}")
            self._wrapped_rcv_disconnected = True
            return msg

        # wrapped_rcv state 3: not yet consumed
        if getattr(self, "_body", None) is not None:
            # body() was called, we return it even if the client disconnected
            self._wrapped_rcv_consumed = True
            return {
                "type": "http.request",
                "body": self._body,
                "more_body": False,
            }
        elif self._stream_consumed:
            # stream() was called to completion
            # return an empty body so that downstream apps don't hang
            # waiting for a disconnect
            self._wrapped_rcv_consumed = True
            return {
                "type": "http.request",
                "body": b"",
                "more_body": False,
            }
        else:
            # body() was never called and stream() wasn't consumed
            try:
                stream = self.stream()
                chunk = await stream.__anext__()
                self._wrapped_rcv_consumed = self._stream_consumed
                return {
                    "type": "http.request",
                    "body": chunk,
                    "more_body": not self._stream_consumed,
                }
            except ClientDisconnect:
                self._wrapped_rcv_disconnected = True
                return {"type": "http.disconnect"}


class ASGIRequestResponseBridge:
    def __init__(self, app: ASGIApp, dispatch: DispatchFunction) -> None:
        self.app = app
        self.dispatch_func = dispatch

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        request = _CachedRequest(scope, receive)
        response = Response(request=request)

        # Set the context for this request
        ctx = Context(request=request)
        token = current_context.set(ctx)
        try:
            wrapped_receive = request.wrapped_receive
            response_sent = anyio.Event()

            async def call_next(
                *_,
            ) -> Response:  # Why cant this be flexible , i mean ☺️🤷‍♀️
                app_exc: Exception | None = None

                async def receive_or_disconnect() -> Message:
                    if response_sent.is_set():
                        return {"type": "http.disconnect"}
                    async with anyio.create_task_group() as task_group:

                        async def wrap(
                            func: typing.Callable[[], typing.Awaitable[T]],
                        ) -> T:
                            result = await func()
                            task_group.cancel_scope.cancel()
                            return result

                        task_group.start_soon(wrap, response_sent.wait)
                        message = await wrap(wrapped_receive)
                    if response_sent.is_set():
                        return {"type": "http.disconnect"}
                    return message

                async def send_no_error(message: Message) -> None:
                    try:
                        await send_stream.send(message)
                    except anyio.BrokenResourceError:
                        raise RuntimeError("No response returned")

                async def coro() -> None:
                    nonlocal app_exc
                    with send_stream:
                        try:
                            await self.app(scope, receive_or_disconnect, send_no_error)
                        except Exception as exc:
                            app_exc = exc

                task_group.start_soon(coro)
                try:
                    message = await recv_stream.receive()
                    info = message.get("info", None)
                    if message["type"] == "http.response.debug" and info is not None:
                        message = await recv_stream.receive()
                except anyio.EndOfStream:
                    if app_exc is not None:
                        raise app_exc
                    raise RuntimeError("Client disconnected before response was sent")
                assert message["type"] == "http.response.start"

                async def body_stream() -> typing.AsyncGenerator[bytes, None]:
                    async for message in recv_stream:
                        assert message["type"] == "http.response.body"
                        body = message.get("body", b"")
                        if body:
                            yield body
                        if not message.get("more_body", False):
                            break
                    if app_exc is not None:
                        raise app_exc

                response_ = response.stream(
                    iterator=body_stream(), status_code=message["status"]
                )  # type: ignore
                response_._response._headers = message["headers"]  # type: ignore
                return response

            streams: anyio.create_memory_object_stream[Message] = (
                anyio.create_memory_object_stream()
            )  # type: ignore
            send_stream, recv_stream = streams
            with recv_stream, send_stream, collapse_excgroups():
                async with anyio.create_task_group() as task_group:
                    await self.dispatch_func(request, response, call_next)  # type: ignore
                    await response.get_response()(scope, wrapped_receive, send)
                    response_sent.set()
                    recv_stream.close()
        finally:
            current_context.reset(token)


WebSocketDispatchFunction = typing.Callable[
    ["WebSocket", typing.Coroutine[None, None, typing.Any]], typing.Awaitable[None]
]


MiddlewareType = typing.Callable[
    [Request, Response, typing.Awaitable[None]],
    typing.Callable[[], typing.Awaitable[None]],
]


def wrap_middleware(middleware_function: DispatchFunction) -> DefineMiddleware:
    return DefineMiddleware(ASGIRequestResponseBridge, dispatch=middleware_function)



#!/usr/bin/env python3
"""
Helper script to set up commitizen and test changelog generation.
"""

import subprocess


def run_command(command, check=True):
    """Run a command and return the result."""
    try:
        result = subprocess.run(
            command, shell=True, check=check, capture_output=True, text=True
        )
        return result
    except subprocess.CalledProcessError as e:
        print(f"Error running command '{command}': {e}")
        return e


def main():
    print("Setting up commitizen for Nexios...")

    # Check if commitizen is installed
    result = run_command("cz --version", check=False)
    if result.returncode != 0:
        print("Installing commitizen...")
        run_command("pip install commitizen")

    # Test commitizen configuration
    print("\nTesting commitizen configuration...")
    result = run_command("cz check --rev-range HEAD~5..HEAD", check=False)
    if result.returncode == 0:
        print("✅ Commitizen is properly configured!")
        print("Recent commits that would trigger a version bump:")
        print(result.stdout)
    else:
        print("ℹ️  No version bump needed for recent commits")
        print("To test, make a commit with conventional format:")
        print("  git commit -m 'feat: add new feature'")
        print("  git commit -m 'fix: fix bug'")
        print("  git commit -m 'BREAKING CHANGE: major change'")

    # Show current version
    print("\nCurrent version: 2.6.2")
    print("To bump version manually:")
    print("  cz bump --yes")


if __name__ == "__main__":
    main()
